# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from unittest.mock import MagicMock, patch, call

from swh.model.hashutil import hash_to_bytes, hash_to_hex

from swh.web.common import service
from swh.web.common.exc import BadInputExc, NotFoundExc
from swh.web.tests.testcase import WebTestCase


class ServiceTestCase(WebTestCase):

    def setUp(self):
        self.BLAKE2S256_SAMPLE = ('685395c5dc57cada459364f0946d3dd45b'
                                  'ad5fcbabc1048edb44380f1d31d0aa')
        self.BLAKE2S256_SAMPLE_BIN = hash_to_bytes(self.BLAKE2S256_SAMPLE)
        self.SHA1_SAMPLE = '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'
        self.SHA1_SAMPLE_BIN = hash_to_bytes(self.SHA1_SAMPLE)
        self.SHA256_SAMPLE = ('8abb0aa566452620ecce816eecdef4792d77a'
                              '293ad8ea82a4d5ecb4d36f7e560')
        self.SHA256_SAMPLE_BIN = hash_to_bytes(self.SHA256_SAMPLE)
        self.SHA1GIT_SAMPLE = '25d1a2e8f32937b0f498a5ca87f823d8df013c01'
        self.SHA1GIT_SAMPLE_BIN = hash_to_bytes(self.SHA1GIT_SAMPLE)
        self.DIRECTORY_ID = '7834ef7e7c357ce2af928115c6c6a42b7e2a44e6'
        self.DIRECTORY_ID_BIN = hash_to_bytes(self.DIRECTORY_ID)
        self.AUTHOR_ID_BIN = {
            'name': b'author',
            'email': b'author@company.org',
        }
        self.AUTHOR_ID = {
            'name': 'author',
            'email': 'author@company.org',
        }
        self.COMMITTER_ID_BIN = {
            'name': b'committer',
            'email': b'committer@corp.org',
        }
        self.COMMITTER_ID = {
            'name': 'committer',
            'email': 'committer@corp.org',
        }
        self.SAMPLE_DATE_RAW = {
            'timestamp': datetime.datetime(
                2000, 1, 17, 11, 23, 54,
                tzinfo=datetime.timezone.utc,
            ).timestamp(),
            'offset': 0,
            'negative_utc': False,
        }
        self.SAMPLE_DATE = '2000-01-17T11:23:54+00:00'
        self.SAMPLE_MESSAGE_BIN = b'elegant fix for bug 31415957'
        self.SAMPLE_MESSAGE = 'elegant fix for bug 31415957'

        self.SAMPLE_REVISION = {
            'id': self.SHA1_SAMPLE,
            'directory': self.DIRECTORY_ID,
            'author': self.AUTHOR_ID,
            'committer': self.COMMITTER_ID,
            'message': self.SAMPLE_MESSAGE,
            'date': self.SAMPLE_DATE,
            'committer_date': self.SAMPLE_DATE,
            'synthetic': False,
            'type': 'git',
            'parents': [],
            'metadata': {},
            'merge': False
        }
        self.SAMPLE_REVISION_RAW = {
            'id': self.SHA1_SAMPLE_BIN,
            'directory': self.DIRECTORY_ID_BIN,
            'author': self.AUTHOR_ID_BIN,
            'committer': self.COMMITTER_ID_BIN,
            'message': self.SAMPLE_MESSAGE_BIN,
            'date': self.SAMPLE_DATE_RAW,
            'committer_date': self.SAMPLE_DATE_RAW,
            'synthetic': False,
            'type': 'git',
            'parents': [],
            'metadata': [],
        }

        self.SAMPLE_CONTENT = {
            'checksums': {
                'blake2s256': self.BLAKE2S256_SAMPLE,
                'sha1': self.SHA1_SAMPLE,
                'sha256': self.SHA256_SAMPLE,
                'sha1_git': self.SHA1GIT_SAMPLE,
            },
            'length': 190,
            'status': 'absent'
        }
        self.SAMPLE_CONTENT_RAW = {
            'blake2s256': self.BLAKE2S256_SAMPLE_BIN,
            'sha1': self.SHA1_SAMPLE_BIN,
            'sha256': self.SHA256_SAMPLE_BIN,
            'sha1_git': self.SHA1GIT_SAMPLE_BIN,
            'length': 190,
            'status': 'hidden'
        }

        self.date_origin_visit1 = datetime.datetime(
            2015, 1, 1, 22, 0, 0,
            tzinfo=datetime.timezone.utc)

        self.origin_visit1 = {
            'date': self.date_origin_visit1,
            'origin': 1,
            'visit': 1
        }

    @patch('swh.web.common.service.storage')
    def test_lookup_multiple_hashes_ball_missing(self, mock_storage):
        # given
        mock_storage.content_missing_per_sha1 = MagicMock(return_value=[])

        # when
        actual_lookup = service.lookup_multiple_hashes(
            [{'filename': 'a',
              'sha1': '456caf10e9535160d90e874b45aa426de762f19f'},
             {'filename': 'b',
              'sha1': '745bab676c8f3cec8016e0c39ea61cf57e518865'}])

        # then
        self.assertEqual(actual_lookup, [
            {'filename': 'a',
             'sha1': '456caf10e9535160d90e874b45aa426de762f19f',
             'found': True},
            {'filename': 'b',
             'sha1': '745bab676c8f3cec8016e0c39ea61cf57e518865',
             'found': True}
        ])

    @patch('swh.web.common.service.storage')
    def test_lookup_multiple_hashes_some_missing(self, mock_storage):
        # given
        mock_storage.content_missing_per_sha1 = MagicMock(return_value=[
            hash_to_bytes('456caf10e9535160d90e874b45aa426de762f19f')
        ])

        # when
        actual_lookup = service.lookup_multiple_hashes(
            [{'filename': 'a',
              'sha1': '456caf10e9535160d90e874b45aa426de762f19f'},
             {'filename': 'b',
              'sha1': '745bab676c8f3cec8016e0c39ea61cf57e518865'}])

        # then
        self.assertEqual(actual_lookup, [
            {'filename': 'a',
             'sha1': '456caf10e9535160d90e874b45aa426de762f19f',
             'found': False},
            {'filename': 'b',
             'sha1': '745bab676c8f3cec8016e0c39ea61cf57e518865',
             'found': True}
        ])

    @patch('swh.web.common.service.storage')
    def test_lookup_hash_does_not_exist(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(return_value=None)

        # when
        actual_lookup = service.lookup_hash(
            'sha1_git:123caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual({'found': None,
                          'algo': 'sha1_git'}, actual_lookup)

        # check the function has been called with parameters
        mock_storage.content_find.assert_called_with(
            {'sha1_git':
             hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')})

    @patch('swh.web.common.service.storage')
    def test_lookup_hash_exist(self, mock_storage):
        # given
        stub_content = {
            'sha1': hash_to_bytes(
                '456caf10e9535160d90e874b45aa426de762f19f')
        }
        mock_storage.content_find = MagicMock(return_value=stub_content)

        # when
        actual_lookup = service.lookup_hash(
            'sha1:456caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual({
            'found': {
                'checksums': {
                    'sha1': '456caf10e9535160d90e874b45aa426de762f19f'
                }
            },
            'algo': 'sha1'
        }, actual_lookup)

        mock_storage.content_find.assert_called_with(
            {'sha1':
             hash_to_bytes('456caf10e9535160d90e874b45aa426de762f19f')}
        )

    @patch('swh.web.common.service.storage')
    def test_search_hash_does_not_exist(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(return_value=None)

        # when
        actual_lookup = service.search_hash(
            'sha1_git:123caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual({'found': False}, actual_lookup)

        # check the function has been called with parameters
        mock_storage.content_find.assert_called_with(
            {'sha1_git':
             hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')})

    @patch('swh.web.common.service.storage')
    def test_search_hash_exist(self, mock_storage):
        # given
        stub_content = {
                'sha1': hash_to_bytes(
                    '456caf10e9535160d90e874b45aa426de762f19f')
            }
        mock_storage.content_find = MagicMock(return_value=stub_content)

        # when
        actual_lookup = service.search_hash(
            'sha1:456caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual({'found': True}, actual_lookup)

        mock_storage.content_find.assert_called_with(
            {'sha1':
             hash_to_bytes('456caf10e9535160d90e874b45aa426de762f19f')},
        )

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_content_ctags(self, mock_idx_storage):
        # given
        mock_idx_storage.content_ctags_get = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'line': 100,
                'name': 'hello',
                'kind': 'function',
                'tool_name': 'ctags',
                'tool_version': 'some-version',
            }])
        expected_ctags = [{
            'id': '123caf10e9535160d90e874b45aa426de762f19f',
            'line': 100,
            'name': 'hello',
            'kind': 'function',
            'tool_name': 'ctags',
            'tool_version': 'some-version',
        }]

        # when
        actual_ctags = list(service.lookup_content_ctags(
            'sha1:123caf10e9535160d90e874b45aa426de762f19f'))

        # then
        self.assertEqual(actual_ctags, expected_ctags)

        mock_idx_storage.content_ctags_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_content_ctags_no_hash(self, mock_idx_storage):
        # given
        mock_idx_storage.content_ctags_get = MagicMock(return_value=[])

        # when
        actual_ctags = list(service.lookup_content_ctags(
            'sha1:123caf10e9535160d90e874b45aa426de762f19f'))

        # then
        self.assertEqual(actual_ctags, [])

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_content_filetype(self, mock_idx_storage):
        # given
        mock_idx_storage.content_mimetype_get = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'mimetype': 'text/x-c++',
                'encoding': 'us-ascii',
            }])
        expected_filetype = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'mimetype': 'text/x-c++',
                'encoding': 'us-ascii',
        }

        # when
        actual_filetype = service.lookup_content_filetype(
            'sha1:123caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_filetype, expected_filetype)

        mock_idx_storage.content_mimetype_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    @patch('swh.web.common.service.storage')
    def test_lookup_content_filetype_2(self, mock_storage, mock_idx_storage):
        # given
        mock_storage.content_find = MagicMock(
            return_value={
                'sha1': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f')
            }
        )
        mock_idx_storage.content_mimetype_get = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'mimetype': 'text/x-python',
                'encoding': 'us-ascii',
            }]
        )
        expected_filetype = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'mimetype': 'text/x-python',
                'encoding': 'us-ascii',
        }

        # when
        actual_filetype = service.lookup_content_filetype(
            'sha1_git:456caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_filetype, expected_filetype)

        mock_storage.content_find(
            'sha1_git', hash_to_bytes(
                '456caf10e9535160d90e874b45aa426de762f19f')
        )
        mock_idx_storage.content_mimetype_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_content_language(self, mock_idx_storage):
        # given
        mock_idx_storage.content_language_get = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'lang': 'python',
            }])
        expected_language = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'lang': 'python',
        }

        # when
        actual_language = service.lookup_content_language(
            'sha1:123caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_language, expected_language)

        mock_idx_storage.content_language_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    @patch('swh.web.common.service.storage')
    def test_lookup_content_language_2(self, mock_storage, mock_idx_storage):
        # given
        mock_storage.content_find = MagicMock(
            return_value={
                'sha1': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f')
            }
        )
        mock_idx_storage.content_language_get = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'lang': 'haskell',
            }]
        )
        expected_language = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'lang': 'haskell',
        }

        # when
        actual_language = service.lookup_content_language(
            'sha1_git:456caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_language, expected_language)

        mock_storage.content_find(
            'sha1_git', hash_to_bytes(
                '456caf10e9535160d90e874b45aa426de762f19f')
        )
        mock_idx_storage.content_language_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_expression(self, mock_idx_storage):
        # given
        mock_idx_storage.content_ctags_search = MagicMock(
            return_value=[{
                'id': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f'),
                'name': 'foobar',
                'kind': 'variable',
                'lang': 'C',
                'line': 10
            }])
        expected_ctags = [{
            'sha1': '123caf10e9535160d90e874b45aa426de762f19f',
            'name': 'foobar',
            'kind': 'variable',
            'lang': 'C',
            'line': 10
        }]

        # when
        actual_ctags = list(service.lookup_expression(
            'foobar', last_sha1='hash', per_page=10))

        # then
        self.assertEqual(actual_ctags, expected_ctags)

        mock_idx_storage.content_ctags_search.assert_called_with(
            'foobar', last_sha1='hash', limit=10)

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_expression_no_result(self, mock_idx_storage):
        # given
        mock_idx_storage.content_ctags_search = MagicMock(
            return_value=[])
        expected_ctags = []

        # when
        actual_ctags = list(service.lookup_expression(
            'barfoo', last_sha1='hash', per_page=10))

        # then
        self.assertEqual(actual_ctags, expected_ctags)

        mock_idx_storage.content_ctags_search.assert_called_with(
            'barfoo', last_sha1='hash', limit=10)

    @patch('swh.web.common.service.idx_storage')
    def test_lookup_content_license(self, mock_idx_storage):
        # given
        mock_idx_storage.content_fossology_license_get = MagicMock(
            return_value=[{
                hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f'): [{
                    'licenses': ['GPL-3.0+'],
                    'tool': {}
                }]
            }])
        expected_license = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'facts': [{
                    'licenses': ['GPL-3.0+'],
                    'tool': {}
                }]
        }

        # when
        actual_license = service.lookup_content_license(
            'sha1:123caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_license, expected_license)

        mock_idx_storage.content_fossology_license_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.idx_storage')
    @patch('swh.web.common.service.storage')
    def test_lookup_content_license_2(self, mock_storage, mock_idx_storage):
        # given
        mock_storage.content_find = MagicMock(
            return_value={
                'sha1': hash_to_bytes(
                    '123caf10e9535160d90e874b45aa426de762f19f')
            }
        )
        mock_idx_storage.content_fossology_license_get = MagicMock(
            return_value=[{
                hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f'): [{
                    'licenses': ['BSD-2-Clause'],
                    'tool': {}
                }]

            }]
        )
        expected_license = {
                'id': '123caf10e9535160d90e874b45aa426de762f19f',
                'facts': [{
                    'licenses': ['BSD-2-Clause'],
                    'tool': {}
                }]
        }

        # when
        actual_license = service.lookup_content_license(
            'sha1_git:456caf10e9535160d90e874b45aa426de762f19f')

        # then
        self.assertEqual(actual_license, expected_license)

        mock_storage.content_find(
            'sha1_git', hash_to_bytes(
                '456caf10e9535160d90e874b45aa426de762f19f')
        )
        mock_idx_storage.content_fossology_license_get.assert_called_with(
            [hash_to_bytes('123caf10e9535160d90e874b45aa426de762f19f')])

    @patch('swh.web.common.service.storage')
    def test_stat_counters(self, mock_storage):
        # given
        input_stats = {
            "content": 1770830,
            "directory": 211683,
            "directory_entry_dir": 209167,
            "directory_entry_file": 1807094,
            "directory_entry_rev": 0,
            "origin": 1096,
            "person": 0,
            "release": 8584,
            "revision": 7792,
            "revision_history": 0,
            "skipped_content": 0
        }
        mock_storage.stat_counters = MagicMock(return_value=input_stats)

        # when
        actual_stats = service.stat_counters()

        # then
        expected_stats = input_stats
        self.assertEqual(actual_stats, expected_stats)

        mock_storage.stat_counters.assert_called_with()

    @patch('swh.web.common.service._lookup_origin_visits')
    def test_lookup_origin_visits(self, mock_lookup_visits):
        # given
        date_origin_visit2 = datetime.datetime(
            2013, 7, 1, 20, 0, 0,
            tzinfo=datetime.timezone.utc)

        date_origin_visit3 = datetime.datetime(
            2015, 1, 1, 21, 0, 0,
            tzinfo=datetime.timezone.utc)
        stub_result = [self.origin_visit1, {
            'date': date_origin_visit2,
            'origin': 1,
            'visit': 2,
            'target': hash_to_bytes(
                '65a55bbdf3629f916219feb3dcc7393ded1bc8db'),
            'branch': b'master',
            'target_type': 'release',
            'metadata': None,
        }, {
            'date': date_origin_visit3,
            'origin': 1,
            'visit': 3
        }]
        mock_lookup_visits.return_value = stub_result

        # when
        expected_origin_visits = [{
            'date': self.origin_visit1['date'].isoformat(),
            'origin': self.origin_visit1['origin'],
            'visit': self.origin_visit1['visit']
        }, {
            'date': date_origin_visit2.isoformat(),
            'origin': 1,
            'visit': 2,
            'target': '65a55bbdf3629f916219feb3dcc7393ded1bc8db',
            'branch': 'master',
            'target_type': 'release',
            'metadata': {},
        }, {
            'date': date_origin_visit3.isoformat(),
            'origin': 1,
            'visit': 3
        }]

        actual_origin_visits = service.lookup_origin_visits(6)

        # then
        self.assertEqual(list(actual_origin_visits), expected_origin_visits)

        mock_lookup_visits.assert_called_once_with(
            6, last_visit=None, limit=10)

    @patch('swh.web.common.service.storage')
    def test_lookup_origin_visit(self, mock_storage):
        # given
        stub_result = self.origin_visit1
        mock_storage.origin_visit_get_by.return_value = stub_result

        expected_origin_visit = {
            'date': self.origin_visit1['date'].isoformat(),
            'origin': self.origin_visit1['origin'],
            'visit': self.origin_visit1['visit']
        }

        # when
        actual_origin_visit = service.lookup_origin_visit(1, 1)

        # then
        self.assertEqual(actual_origin_visit, expected_origin_visit)

        mock_storage.origin_visit_get_by.assert_called_once_with(1, 1)

    @patch('swh.web.common.service.storage')
    def test_lookup_origin(self, mock_storage):
        # given
        mock_storage.origin_get = MagicMock(return_value={
            'id': 'origin-id',
            'url': 'ftp://some/url/to/origin',
            'type': 'ftp'})

        # when
        actual_origin = service.lookup_origin({'id': 'origin-id'})

        # then
        self.assertEqual(actual_origin, {'id': 'origin-id',
                                         'url': 'ftp://some/url/to/origin',
                                         'type': 'ftp'})

        mock_storage.origin_get.assert_called_with({'id': 'origin-id'})

    @patch('swh.web.common.service.storage')
    def test_lookup_release_ko_id_checksum_not_a_sha1(self, mock_storage):
        # given
        mock_storage.release_get = MagicMock()

        with self.assertRaises(BadInputExc) as cm:
            # when
            service.lookup_release('not-a-sha1')
        self.assertIn('invalid checksum', cm.exception.args[0].lower())

        mock_storage.release_get.called = False

    @patch('swh.web.common.service.storage')
    def test_lookup_release_ko_id_checksum_too_long(self, mock_storage):
        # given
        mock_storage.release_get = MagicMock()

        # when
        with self.assertRaises(BadInputExc) as cm:
            service.lookup_release(
                '13c1d34d138ec13b5ebad226dc2528dc7506c956e4646f62d4daf5'
                '1aea892abe')
        self.assertEqual('Only sha1_git is supported.', cm.exception.args[0])

        mock_storage.release_get.called = False

    @patch('swh.web.common.service.storage')
    def test_lookup_directory_with_path_not_found(self, mock_storage):
        # given
        mock_storage.lookup_directory_with_path = MagicMock(return_value=None)

        sha1_git = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'

        # when
        actual_directory = mock_storage.lookup_directory_with_path(
            sha1_git, 'some/path/here')

        self.assertIsNone(actual_directory)

    @patch('swh.web.common.service.storage')
    def test_lookup_directory_with_path_found(self, mock_storage):
        # given
        sha1_git = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'
        entry = {'id': 'dir-id',
                 'type': 'dir',
                 'name': 'some/path/foo'}

        mock_storage.lookup_directory_with_path = MagicMock(return_value=entry)

        # when
        actual_directory = mock_storage.lookup_directory_with_path(
            sha1_git, 'some/path/here')

        self.assertEqual(entry, actual_directory)

    @patch('swh.web.common.service.storage')
    def test_lookup_release(self, mock_storage):
        # given
        mock_storage.release_get = MagicMock(return_value=[{
            'id': hash_to_bytes('65a55bbdf3629f916219feb3dcc7393ded1bc8db'),
            'target': None,
            'date': {
                'timestamp': datetime.datetime(
                    2015, 1, 1, 22, 0, 0,
                    tzinfo=datetime.timezone.utc).timestamp(),
                'offset': 0,
                'negative_utc': True,
            },
            'name': b'v0.0.1',
            'message': b'synthetic release',
            'synthetic': True,
        }])

        # when
        actual_release = service.lookup_release(
            '65a55bbdf3629f916219feb3dcc7393ded1bc8db')

        # then
        self.assertEqual(actual_release, {
            'id': '65a55bbdf3629f916219feb3dcc7393ded1bc8db',
            'target': None,
            'date': '2015-01-01T22:00:00-00:00',
            'name': 'v0.0.1',
            'message': 'synthetic release',
            'synthetic': True,
        })

        mock_storage.release_get.assert_called_with(
            [hash_to_bytes('65a55bbdf3629f916219feb3dcc7393ded1bc8db')])

    def test_lookup_revision_with_context_ko_not_a_sha1_1(self):
        # given
        sha1_git = '13c1d34d138ec13b5ebad226dc2528dc7506c956e4646f62d4' \
                   'daf51aea892abe'
        sha1_git_root = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'

        # when
        with self.assertRaises(BadInputExc) as cm:
            service.lookup_revision_with_context(sha1_git_root, sha1_git)
        self.assertIn('Only sha1_git is supported', cm.exception.args[0])

    def test_lookup_revision_with_context_ko_not_a_sha1_2(self):
        # given
        sha1_git_root = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'
        sha1_git = '13c1d34d138ec13b5ebad226dc2528dc7506c956e4646f6' \
                   '2d4daf51aea892abe'

        # when
        with self.assertRaises(BadInputExc) as cm:
            service.lookup_revision_with_context(sha1_git_root, sha1_git)
        self.assertIn('Only sha1_git is supported', cm.exception.args[0])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_with_context_ko_sha1_git_does_not_exist(
            self,
            mock_storage):
        # given
        sha1_git_root = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'
        sha1_git = '777777bdf3629f916219feb3dcc7393ded1bc8db'

        sha1_git_bin = hash_to_bytes(sha1_git)

        mock_storage.revision_get.return_value = None

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_revision_with_context(sha1_git_root, sha1_git)
        self.assertIn('Revision 777777bdf3629f916219feb3dcc7393ded1bc8db'
                      ' not found', cm.exception.args[0])

        mock_storage.revision_get.assert_called_once_with(
            [sha1_git_bin])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_with_context_ko_root_sha1_git_does_not_exist(
            self,
            mock_storage):
        # given
        sha1_git_root = '65a55bbdf3629f916219feb3dcc7393ded1bc8db'
        sha1_git = '777777bdf3629f916219feb3dcc7393ded1bc8db'

        sha1_git_root_bin = hash_to_bytes(sha1_git_root)
        sha1_git_bin = hash_to_bytes(sha1_git)

        mock_storage.revision_get.side_effect = ['foo', None]

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_revision_with_context(sha1_git_root, sha1_git)
        self.assertIn('Revision root 65a55bbdf3629f916219feb3dcc7393ded1bc8db'
                      ' not found', cm.exception.args[0])

        mock_storage.revision_get.assert_has_calls([call([sha1_git_bin]),
                                                    call([sha1_git_root_bin])])

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_revision_with_context(self, mock_query, mock_storage):
        # given
        sha1_git_root = '666'
        sha1_git = '883'

        sha1_git_root_bin = b'666'
        sha1_git_bin = b'883'

        sha1_git_root_dict = {
            'id': sha1_git_root_bin,
            'parents': [b'999'],
        }
        sha1_git_dict = {
            'id': sha1_git_bin,
            'parents': [],
            'directory': b'278',
        }

        stub_revisions = [
            sha1_git_root_dict,
            {
                'id': b'999',
                'parents': [b'777', b'883', b'888'],
            },
            {
                'id': b'777',
                'parents': [b'883'],
            },
            sha1_git_dict,
            {
                'id': b'888',
                'parents': [b'889'],
            },
            {
                'id': b'889',
                'parents': [],
            },
        ]

        # inputs ok
        mock_query.parse_hash_with_algorithms_or_throws.side_effect = [
            ('sha1', sha1_git_bin),
            ('sha1', sha1_git_root_bin)
        ]

        # lookup revision first 883, then 666 (both exists)
        mock_storage.revision_get.return_value = [
            sha1_git_dict,
            sha1_git_root_dict
        ]

        mock_storage.revision_log = MagicMock(
            return_value=stub_revisions)

        # when

        actual_revision = service.lookup_revision_with_context(
            sha1_git_root,
            sha1_git)

        # then
        self.assertEqual(actual_revision, {
            'id': hash_to_hex(sha1_git_bin),
            'parents': [],
            'children': [hash_to_hex(b'999'), hash_to_hex(b'777')],
            'directory': hash_to_hex(b'278'),
            'merge': False
        })

        mock_query.parse_hash_with_algorithms_or_throws.assert_has_calls(
            [call(sha1_git, ['sha1'], 'Only sha1_git is supported.'),
             call(sha1_git_root, ['sha1'], 'Only sha1_git is supported.')])

        mock_storage.revision_log.assert_called_with(
            [sha1_git_root_bin], 100)

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_revision_with_context_retrieved_as_dict(
            self, mock_query, mock_storage):
        # given
        sha1_git = '883'

        sha1_git_root_bin = b'666'
        sha1_git_bin = b'883'

        sha1_git_root_dict = {
            'id': sha1_git_root_bin,
            'parents': [b'999'],
        }

        sha1_git_dict = {
            'id': sha1_git_bin,
            'parents': [],
            'directory': b'278',
        }

        stub_revisions = [
            sha1_git_root_dict,
            {
                'id': b'999',
                'parents': [b'777', b'883', b'888'],
            },
            {
                'id': b'777',
                'parents': [b'883'],
            },
            sha1_git_dict,
            {
                'id': b'888',
                'parents': [b'889'],
            },
            {
                'id': b'889',
                'parents': [],
            },
        ]

        # inputs ok
        mock_query.parse_hash_with_algorithms_or_throws.return_value = (
            'sha1', sha1_git_bin)

        # lookup only on sha1
        mock_storage.revision_get.return_value = [sha1_git_dict]

        mock_storage.revision_log.return_value = stub_revisions

        # when
        actual_revision = service.lookup_revision_with_context(
            {'id': sha1_git_root_bin},
            sha1_git)

        # then
        self.assertEqual(actual_revision, {
            'id': hash_to_hex(sha1_git_bin),
            'parents': [],
            'children': [hash_to_hex(b'999'), hash_to_hex(b'777')],
            'directory': hash_to_hex(b'278'),
            'merge': False
        })

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with(  # noqa
            sha1_git, ['sha1'], 'Only sha1_git is supported.')

        mock_storage.revision_get.assert_called_once_with([sha1_git_bin])

        mock_storage.revision_log.assert_called_with(
            [sha1_git_root_bin], 100)

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_not_found(self,
                                                      mock_query,
                                                      mock_storage):
        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')
        mock_storage.revision_get.return_value = None

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_directory_with_revision('123')
        self.assertIn('Revision 123 not found', cm.exception.args[0])

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_ko_revision_with_path_to_nowhere(
            self,
            mock_query,
            mock_storage):
        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        mock_storage.directory_entry_get_by_path.return_value = None

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_directory_with_revision(
                '123',
                'path/to/something/unknown')
        exception_text = cm.exception.args[0].lower()
        self.assertIn('directory or file', exception_text)
        self.assertIn('path/to/something/unknown', exception_text)
        self.assertIn('revision 123', exception_text)
        self.assertIn('not found', exception_text)

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_entry_get_by_path.assert_called_once_with(
            b'dir-id-as-sha1', [b'path', b'to', b'something', b'unknown'])

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_ko_type_not_implemented(
            self,
            mock_query,
            mock_storage):

        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        mock_storage.directory_entry_get_by_path.return_value = {
            'type': 'rev',
            'name': b'some/path/to/rev',
            'target': b'456'
        }

        stub_content = {
            'id': b'12',
            'type': 'file'
        }

        mock_storage.content_get.return_value = stub_content

        # when
        with self.assertRaises(NotImplementedError) as cm:
            service.lookup_directory_with_revision(
                '123',
                'some/path/to/rev')
        self.assertIn("Entity of type rev not implemented.",
                      cm.exception.args[0])

        # then
        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_entry_get_by_path.assert_called_once_with(
            b'dir-id-as-sha1', [b'some', b'path', b'to', b'rev'])

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_revision_without_path(
        self, mock_query, mock_storage,
    ):
        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        stub_dir_entries = [{
            'id': b'123',
            'type': 'dir'
        }, {
            'id': b'456',
            'type': 'file'
        }]

        mock_storage.directory_ls.return_value = stub_dir_entries

        # when
        actual_directory_entries = service.lookup_directory_with_revision(
            '123')

        self.assertEqual(actual_directory_entries['type'], 'dir')
        self.assertEqual(list(actual_directory_entries['content']),
                         stub_dir_entries)

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_ls.assert_called_once_with(dir_id)

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_with_path_to_dir(self,
                                                             mock_query,
                                                             mock_storage):
        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        stub_dir_entries = [{
            'id': b'12',
            'type': 'dir'
        }, {
            'id': b'34',
            'type': 'file'
        }]

        mock_storage.directory_entry_get_by_path.return_value = {
            'type': 'dir',
            'name': b'some/path',
            'target': b'456'
        }
        mock_storage.directory_ls.return_value = stub_dir_entries

        # when
        actual_directory_entries = service.lookup_directory_with_revision(
            '123',
            'some/path')

        self.assertEqual(actual_directory_entries['type'], 'dir')
        self.assertEqual(actual_directory_entries['revision'], '123')
        self.assertEqual(actual_directory_entries['path'], 'some/path')
        self.assertEqual(list(actual_directory_entries['content']),
                         stub_dir_entries)

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_entry_get_by_path.assert_called_once_with(
            dir_id,
            [b'some', b'path'])
        mock_storage.directory_ls.assert_called_once_with(b'456')

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_with_path_to_file_wo_data(
            self,
            mock_query,
            mock_storage):

        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        mock_storage.directory_entry_get_by_path.return_value = {
                'type': 'file',
                'name': b'some/path/to/file',
                'target': b'789'
            }

        stub_content = {
            'status': 'visible',
        }

        mock_storage.content_find.return_value = stub_content

        # when
        actual_content = service.lookup_directory_with_revision(
            '123',
            'some/path/to/file')

        # then
        self.assertEqual(actual_content, {'type': 'file',
                                          'revision': '123',
                                          'path': 'some/path/to/file',
                                          'content': stub_content})

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_entry_get_by_path.assert_called_once_with(
            b'dir-id-as-sha1', [b'some', b'path', b'to', b'file'])
        mock_storage.content_find.assert_called_once_with({'sha1_git': b'789'})

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_with_revision_with_path_to_file_w_data(
            self,
            mock_query,
            mock_storage):

        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = ('sha1',
                                                                        b'123')

        dir_id = b'dir-id-as-sha1'
        mock_storage.revision_get.return_value = [{
            'directory': dir_id,
        }]

        mock_storage.directory_entry_get_by_path.return_value = {
                'type': 'file',
                'name': b'some/path/to/file',
                'target': b'789'
            }

        stub_content = {
            'status': 'visible',
            'sha1': b'content-sha1'
        }

        mock_storage.content_find.return_value = stub_content
        mock_storage.content_get.return_value = [{
            'sha1': b'content-sha1',
            'data': b'some raw data'
        }]

        expected_content = {
            'status': 'visible',
            'checksums': {
                'sha1': hash_to_hex(b'content-sha1'),
            },
            'data': b'some raw data'
        }

        # when
        actual_content = service.lookup_directory_with_revision(
            '123',
            'some/path/to/file',
            with_data=True)

        # then
        self.assertEqual(actual_content, {'type': 'file',
                                          'revision': '123',
                                          'path': 'some/path/to/file',
                                          'content': expected_content})

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_once_with
        ('123', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.revision_get.assert_called_once_with([b'123'])
        mock_storage.directory_entry_get_by_path.assert_called_once_with(
            b'dir-id-as-sha1', [b'some', b'path', b'to', b'file'])
        mock_storage.content_find.assert_called_once_with({'sha1_git': b'789'})
        mock_storage.content_get.assert_called_once_with([b'content-sha1'])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision(self, mock_storage):
        # given
        mock_storage.revision_get = MagicMock(
            return_value=[self.SAMPLE_REVISION_RAW])

        # when
        actual_revision = service.lookup_revision(
            self.SHA1_SAMPLE)

        # then
        self.assertEqual(actual_revision, self.SAMPLE_REVISION)

        mock_storage.revision_get.assert_called_with(
            [self.SHA1_SAMPLE_BIN])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_invalid_msg(self, mock_storage):
        # given
        stub_rev = self.SAMPLE_REVISION_RAW
        stub_rev['message'] = b'elegant fix for bug \xff'

        expected_revision = self.SAMPLE_REVISION
        expected_revision['message'] = None
        expected_revision['message_decoding_failed'] = True
        mock_storage.revision_get = MagicMock(return_value=[stub_rev])

        # when
        actual_revision = service.lookup_revision(
            self.SHA1_SAMPLE)

        # then
        self.assertEqual(actual_revision, expected_revision)

        mock_storage.revision_get.assert_called_with(
            [self.SHA1_SAMPLE_BIN])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_msg_ok(self, mock_storage):
        # given
        mock_storage.revision_get.return_value = [self.SAMPLE_REVISION_RAW]

        # when
        rv = service.lookup_revision_message(
            self.SHA1_SAMPLE)

        # then
        self.assertEqual(rv, {'message': self.SAMPLE_MESSAGE_BIN})
        mock_storage.revision_get.assert_called_with(
            [self.SHA1_SAMPLE_BIN])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_msg_absent(self, mock_storage):
        # given
        stub_revision = self.SAMPLE_REVISION_RAW
        del stub_revision['message']
        mock_storage.revision_get.return_value = stub_revision

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_revision_message(
                self.SHA1_SAMPLE)

        # then
        mock_storage.revision_get.assert_called_with(
            [self.SHA1_SAMPLE_BIN])
        self.assertEqual(
            cm.exception.args[0],
            'No message for revision with sha1_git %s.' % self.SHA1_SAMPLE,
        )

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_msg_norev(self, mock_storage):
        # given
        mock_storage.revision_get.return_value = None

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_revision_message(
                self.SHA1_SAMPLE)

        # then
        mock_storage.revision_get.assert_called_with(
            [self.SHA1_SAMPLE_BIN])
        self.assertEqual(
            cm.exception.args[0],
            'Revision with sha1_git %s not found.' % self.SHA1_SAMPLE,
        )

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_multiple(self, mock_storage):
        # given
        sha1 = self.SHA1_SAMPLE
        sha1_other = 'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc'

        stub_revisions = [
            self.SAMPLE_REVISION_RAW,
            {
                'id': hash_to_bytes(sha1_other),
                'directory': 'abcdbe353ed3480476f032475e7c233eff7371d5',
                'author': {
                    'name': b'name',
                    'email': b'name@surname.org',
                },
                'committer': {
                    'name': b'name',
                    'email': b'name@surname.org',
                },
                'message': b'ugly fix for bug 42',
                'date': {
                    'timestamp': datetime.datetime(
                        2000, 1, 12, 5, 23, 54,
                        tzinfo=datetime.timezone.utc).timestamp(),
                    'offset': 0,
                    'negative_utc': False
                    },
                'date_offset': 0,
                'committer_date': {
                    'timestamp': datetime.datetime(
                        2000, 1, 12, 5, 23, 54,
                        tzinfo=datetime.timezone.utc).timestamp(),
                    'offset': 0,
                    'negative_utc': False
                    },
                'committer_date_offset': 0,
                'synthetic': False,
                'type': 'git',
                'parents': [],
                'metadata': [],
            }
        ]

        mock_storage.revision_get.return_value = stub_revisions

        # when
        actual_revisions = service.lookup_revision_multiple(
            [sha1, sha1_other])

        # then
        self.assertEqual(list(actual_revisions), [
            self.SAMPLE_REVISION,
            {
                'id': sha1_other,
                'directory': 'abcdbe353ed3480476f032475e7c233eff7371d5',
                'author': {
                    'name': 'name',
                    'email': 'name@surname.org',
                },
                'committer': {
                    'name': 'name',
                    'email': 'name@surname.org',
                },
                'message': 'ugly fix for bug 42',
                'date': '2000-01-12T05:23:54+00:00',
                'date_offset': 0,
                'committer_date': '2000-01-12T05:23:54+00:00',
                'committer_date_offset': 0,
                'synthetic': False,
                'type': 'git',
                'parents': [],
                'metadata': {},
                'merge': False
            }
        ])

        self.assertEqual(
            list(mock_storage.revision_get.call_args[0][0]),
            [hash_to_bytes(sha1),
             hash_to_bytes(sha1_other)])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_multiple_none_found(self, mock_storage):
        # given
        sha1_bin = self.SHA1_SAMPLE
        sha1_other = 'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc'

        mock_storage.revision_get.return_value = []

        # then
        actual_revisions = service.lookup_revision_multiple(
            [sha1_bin, sha1_other])

        self.assertEqual(list(actual_revisions), [])

        self.assertEqual(
            list(mock_storage.revision_get.call_args[0][0]),
            [hash_to_bytes(self.SHA1_SAMPLE),
             hash_to_bytes(sha1_other)])

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_log(self, mock_storage):
        # given
        stub_revision_log = [self.SAMPLE_REVISION_RAW]
        mock_storage.revision_log = MagicMock(return_value=stub_revision_log)

        # when
        actual_revision = service.lookup_revision_log(
            'abcdbe353ed3480476f032475e7c233eff7371d5',
            limit=25)

        # then
        self.assertEqual(list(actual_revision), [self.SAMPLE_REVISION])

        mock_storage.revision_log.assert_called_with(
            [hash_to_bytes('abcdbe353ed3480476f032475e7c233eff7371d5')], 25)

    @patch('swh.web.common.service.lookup_revision_log')
    @patch('swh.web.common.service.lookup_snapshot')
    @patch('swh.web.common.service.get_origin_visit')
    def test_lookup_revision_log_by(self, mock_get_origin_visit,
                                    mock_lookup_snapshot,
                                    mock_lookup_revision_log):
        # given
        mock_get_origin_visit.return_value = {'snapshot': self.SHA1_SAMPLE}
        mock_lookup_snapshot.return_value = \
            {
                'branches': {
                    'refs/heads/master': {
                        'target_type': 'revision',
                        'target': self.SAMPLE_REVISION['id']
                    }
                }
            }

        mock_lookup_revision_log.return_value = [self.SAMPLE_REVISION]

        # when
        actual_log = service.lookup_revision_log_by(
            1, 'refs/heads/master', None, limit=100)
        # then
        self.assertEqual(list(actual_log), [self.SAMPLE_REVISION])

    @patch('swh.web.common.service.lookup_snapshot')
    @patch('swh.web.common.service.get_origin_visit')
    def test_lookup_revision_log_by_notfound(self, mock_get_origin_visit,
                                             mock_lookup_snapshot):
        # given
        mock_get_origin_visit.return_value = {'snapshot': self.SHA1_SAMPLE}
        mock_lookup_snapshot.return_value = {'branches': {}}

        # when
        with self.assertRaises(NotFoundExc):
            service.lookup_revision_log_by(
                1, 'refs/heads/master', None, limit=100)

    @patch('swh.web.common.service.storage')
    def test_lookup_content_raw_not_found(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(return_value=None)

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_content_raw('sha1:' + self.SHA1_SAMPLE)
        self.assertIn(cm.exception.args[0],
                      'Content with %s checksum equals to %s not found!' %
                      ('sha1', self.SHA1_SAMPLE))

        mock_storage.content_find.assert_called_with(
            {'sha1': hash_to_bytes(self.SHA1_SAMPLE)})

    @patch('swh.web.common.service.storage')
    def test_lookup_content_raw(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(return_value={
            'sha1': self.SHA1_SAMPLE,
        })
        mock_storage.content_get = MagicMock(return_value=[{
            'data': b'binary data'}])

        # when
        actual_content = service.lookup_content_raw(
            'sha256:%s' % self.SHA256_SAMPLE)

        # then
        self.assertEqual(actual_content, {'data': b'binary data'})

        mock_storage.content_find.assert_called_once_with(
            {'sha256': self.SHA256_SAMPLE_BIN})
        mock_storage.content_get.assert_called_once_with(
            [hash_to_bytes(self.SHA1_SAMPLE)])

    @patch('swh.web.common.service.storage')
    def test_lookup_content_not_found(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(return_value=None)

        # when
        with self.assertRaises(NotFoundExc) as cm:
            # then
            service.lookup_content('sha1:%s' % self.SHA1_SAMPLE)
        self.assertIn(cm.exception.args[0],
                      'Content with %s checksum equals to %s not found!' %
                      ('sha1', self.SHA1_SAMPLE))

        mock_storage.content_find.assert_called_with(
            {'sha1': self.SHA1_SAMPLE_BIN})

    @patch('swh.web.common.service.storage')
    def test_lookup_content_with_sha1(self, mock_storage):
        # given
        mock_storage.content_find = MagicMock(
            return_value=self.SAMPLE_CONTENT_RAW)

        # when
        actual_content = service.lookup_content(
            'sha1:%s' % self.SHA1_SAMPLE)

        # then
        self.assertEqual(actual_content, self.SAMPLE_CONTENT)

        mock_storage.content_find.assert_called_with(
            {'sha1': hash_to_bytes(self.SHA1_SAMPLE)})

    @patch('swh.web.common.service.storage')
    def test_lookup_content_with_sha256(self, mock_storage):
        # given
        stub_content = self.SAMPLE_CONTENT_RAW
        stub_content['status'] = 'visible'

        expected_content = self.SAMPLE_CONTENT
        expected_content['status'] = 'visible'
        mock_storage.content_find = MagicMock(
            return_value=stub_content)

        # when
        actual_content = service.lookup_content(
            'sha256:%s' % self.SHA256_SAMPLE)

        # then
        self.assertEqual(actual_content, expected_content)

        mock_storage.content_find.assert_called_with(
            {'sha256': self.SHA256_SAMPLE_BIN})

    @patch('swh.web.common.service.storage')
    def test_lookup_person(self, mock_storage):
        # given
        mock_storage.person_get = MagicMock(return_value=[{
            'id': 'person_id',
            'name': b'some_name',
            'email': b'some-email',
        }])

        # when
        actual_person = service.lookup_person('person_id')

        # then
        self.assertEqual(actual_person, {
            'id': 'person_id',
            'name': 'some_name',
            'email': 'some-email',
        })

        mock_storage.person_get.assert_called_with(['person_id'])

    @patch('swh.web.common.service.storage')
    def test_lookup_directory_bad_checksum(self, mock_storage):
        # given
        mock_storage.directory_ls = MagicMock()

        # when
        with self.assertRaises(BadInputExc):
            service.lookup_directory('directory_id')

        # then
        mock_storage.directory_ls.called = False

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory_not_found(self, mock_query, mock_storage):
        # given
        mock_query.parse_hash_with_algorithms_or_throws.return_value = (
            'sha1',
            'directory-id-bin')
        mock_storage.directory_missing.return_value = ['directory-id-bin']

        # when
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_directory('directory_id')

        self.assertIn('Directory with sha1_git directory_id not found',
                      cm.exception.args[0])

        # then
        mock_query.parse_hash_with_algorithms_or_throws.assert_called_with(
            'directory_id', ['sha1'], 'Only sha1_git is supported.')

    @patch('swh.web.common.service.storage')
    @patch('swh.web.common.service.query')
    def test_lookup_directory(self, mock_query, mock_storage):
        mock_query.parse_hash_with_algorithms_or_throws.return_value = (
            'sha1',
            'directory-sha1-bin')

        # given
        stub_dir_entries = [{
            'sha1': self.SHA1_SAMPLE_BIN,
            'sha256': self.SHA256_SAMPLE_BIN,
            'sha1_git': self.SHA1GIT_SAMPLE_BIN,
            'blake2s256': self.BLAKE2S256_SAMPLE_BIN,
            'target': hash_to_bytes(
                '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03'),
            'dir_id': self.DIRECTORY_ID_BIN,
            'name': b'bob',
            'type': 10,
        }]

        expected_dir_entries = [{
            'checksums': {
                'sha1': self.SHA1_SAMPLE,
                'sha256': self.SHA256_SAMPLE,
                'sha1_git': self.SHA1GIT_SAMPLE,
                'blake2s256': self.BLAKE2S256_SAMPLE
            },
            'target': '40e71b8614fcd89ccd17ca2b1d9e66c5b00a6d03',
            'dir_id': self.DIRECTORY_ID,
            'name': 'bob',
            'type': 10,
        }]

        mock_storage.directory_ls.return_value = stub_dir_entries
        mock_storage.directory_missing.return_value = []

        # when
        actual_directory_ls = list(service.lookup_directory(
            'directory-sha1'))

        # then
        self.assertEqual(actual_directory_ls, expected_dir_entries)

        mock_query.parse_hash_with_algorithms_or_throws.assert_called_with(
            'directory-sha1', ['sha1'], 'Only sha1_git is supported.')
        mock_storage.directory_ls.assert_called_with(
            'directory-sha1-bin')

    @patch('swh.web.common.service.storage')
    def test_lookup_directory_empty(self, mock_storage):
        empty_dir_sha1 = '4b825dc642cb6eb9a060e54bf8d69288fbee4904'
        mock_storage.directory_ls.return_value = []

        # when
        actual_directory_ls = list(service.lookup_directory(empty_dir_sha1))

        # then
        self.assertEqual(actual_directory_ls, [])

        self.assertFalse(mock_storage.directory_ls.called)

    @patch('swh.web.common.service.lookup_snapshot')
    @patch('swh.web.common.service.get_origin_visit')
    def test_lookup_revision_by_nothing_found(self, mock_get_origin_visit,
                                              mock_lookup_snapshot):
        # given
        mock_get_origin_visit.return_value = {'snapshot': self.SHA1_SAMPLE}
        mock_lookup_snapshot.return_value = {'branches': {}}

        # when
        with self.assertRaises(NotFoundExc):
            service.lookup_revision_by(1)

    @patch('swh.web.common.service.lookup_revision')
    @patch('swh.web.common.service.lookup_snapshot')
    @patch('swh.web.common.service.get_origin_visit')
    def test_lookup_revision_by(self, mock_get_origin_visit,
                                mock_lookup_snapshot, mock_lookup_revision):
        # given
        expected_rev = self.SAMPLE_REVISION

        mock_get_origin_visit.return_value = {'snapshot': self.SHA1_SAMPLE}
        mock_lookup_snapshot.return_value = \
            {
                'branches': {
                    'master2': {
                        'target_type': 'revision',
                        'target': expected_rev['id']
                    }
                }
            }

        mock_lookup_revision.return_value = expected_rev

        # when
        actual_revision = service.lookup_revision_by(10, 'master2', 'some-ts')

        # then
        self.assertEqual(actual_revision, expected_rev)

    @patch('swh.web.common.service.storage')
    def test_lookup_revision_with_context_by_ko(self, mock_storage):
        # given
        mock_storage.revision_get_by.return_value = None

        # when
        origin_id = 1
        branch_name = 'master3'
        ts = None
        with self.assertRaises(NotFoundExc) as cm:
            service.lookup_revision_with_context_by(origin_id, branch_name, ts,
                                                    'sha1')
        # then
        self.assertIn(
            'Revision with (origin_id: %s, branch_name: %s'
            ', ts: %s) not found.' % (origin_id,
                                      branch_name,
                                      ts), cm.exception.args[0])

        mock_storage.revision_get_by.assert_called_once_with(
            origin_id, branch_name, limit=1, timestamp=ts)

    @patch('swh.web.common.service.lookup_revision_with_context')
    @patch('swh.web.common.service.storage')
    def test_lookup_revision_with_context_by(
            self, mock_storage, mock_lookup_revision_with_context,
    ):
        # given
        stub_root_rev = {'id': 'root-rev-id'}
        mock_storage.revision_get_by.return_value = [{'id': 'root-rev-id'}]
        stub_rev = {'id': 'rev-found'}
        mock_lookup_revision_with_context.return_value = stub_rev

        # when
        origin_id = 1
        branch_name = 'master3'
        ts = None
        sha1_git = 'sha1'
        actual_root_rev, actual_rev = service.lookup_revision_with_context_by(
            origin_id, branch_name, ts, sha1_git)

        # then
        self.assertEqual(actual_root_rev, stub_root_rev)
        self.assertEqual(actual_rev, stub_rev)

        mock_storage.revision_get_by.assert_called_once_with(
            origin_id, branch_name, limit=1, timestamp=ts)
        mock_lookup_revision_with_context.assert_called_once_with(
            stub_root_rev, sha1_git, 100)

    def test_lookup_revision_through_ko_not_implemented(self):
        # then
        with self.assertRaises(NotImplementedError):
            service.lookup_revision_through({
                'something-unknown': 10,
            })

    @patch('swh.web.common.service.lookup_revision_with_context_by')
    def test_lookup_revision_through_with_context_by(self, mock_lookup):
        # given
        stub_rev = {'id': 'rev'}
        mock_lookup.return_value = stub_rev

        # when
        actual_revision = service.lookup_revision_through({
            'origin_id': 1,
            'branch_name': 'master',
            'ts': None,
            'sha1_git': 'sha1-git'
        }, limit=1000)

        # then
        self.assertEqual(actual_revision, stub_rev)

        mock_lookup.assert_called_once_with(
            1, 'master', None, 'sha1-git', 1000)

    @patch('swh.web.common.service.lookup_revision_by')
    def test_lookup_revision_through_with_revision_by(self, mock_lookup):
        # given
        stub_rev = {'id': 'rev'}
        mock_lookup.return_value = stub_rev

        # when
        actual_revision = service.lookup_revision_through({
            'origin_id': 2,
            'branch_name': 'master2',
            'ts': 'some-ts',
        }, limit=10)

        # then
        self.assertEqual(actual_revision, stub_rev)

        mock_lookup.assert_called_once_with(
            2, 'master2', 'some-ts')

    @patch('swh.web.common.service.lookup_revision_with_context')
    def test_lookup_revision_through_with_context(self, mock_lookup):
        # given
        stub_rev = {'id': 'rev'}
        mock_lookup.return_value = stub_rev

        # when
        actual_revision = service.lookup_revision_through({
            'sha1_git_root': 'some-sha1-root',
            'sha1_git': 'some-sha1',
        })

        # then
        self.assertEqual(actual_revision, stub_rev)

        mock_lookup.assert_called_once_with(
            'some-sha1-root', 'some-sha1', 100)

    @patch('swh.web.common.service.lookup_revision')
    def test_lookup_revision_through_with_revision(self, mock_lookup):
        # given
        stub_rev = {'id': 'rev'}
        mock_lookup.return_value = stub_rev

        # when
        actual_revision = service.lookup_revision_through({
            'sha1_git': 'some-sha1',
        })

        # then
        self.assertEqual(actual_revision, stub_rev)

        mock_lookup.assert_called_once_with(
            'some-sha1')

    @patch('swh.web.common.service.lookup_revision_through')
    def test_lookup_directory_through_revision_ko_not_found(
            self, mock_lookup_rev):
        # given
        mock_lookup_rev.return_value = None

        # when
        with self.assertRaises(NotFoundExc):
            service.lookup_directory_through_revision(
                {'id': 'rev'}, 'some/path', 100)

        mock_lookup_rev.assert_called_once_with({'id': 'rev'}, 100)

    @patch('swh.web.common.service.lookup_revision_through')
    @patch('swh.web.common.service.lookup_directory_with_revision')
    def test_lookup_directory_through_revision_ok_with_data(
            self, mock_lookup_dir, mock_lookup_rev):
        # given
        mock_lookup_rev.return_value = {'id': 'rev-id'}
        mock_lookup_dir.return_value = {'type': 'dir',
                                        'content': []}

        # when
        rev_id, dir_result = service.lookup_directory_through_revision(
            {'id': 'rev'}, 'some/path', 100)
        # then
        self.assertEqual(rev_id, 'rev-id')
        self.assertEqual(dir_result, {'type': 'dir',
                                      'content': []})

        mock_lookup_rev.assert_called_once_with({'id': 'rev'}, 100)
        mock_lookup_dir.assert_called_once_with('rev-id', 'some/path', False)

    @patch('swh.web.common.service.lookup_revision_through')
    @patch('swh.web.common.service.lookup_directory_with_revision')
    def test_lookup_directory_through_revision_ok_with_content(
            self, mock_lookup_dir, mock_lookup_rev):
        # given
        mock_lookup_rev.return_value = {'id': 'rev-id'}
        stub_result = {'type': 'file',
                       'revision': 'rev-id',
                       'content': {'data': b'blah',
                                   'sha1': 'sha1'}}
        mock_lookup_dir.return_value = stub_result

        # when
        rev_id, dir_result = service.lookup_directory_through_revision(
            {'id': 'rev'}, 'some/path', 10, with_data=True)
        # then
        self.assertEqual(rev_id, 'rev-id')
        self.assertEqual(dir_result, stub_result)

        mock_lookup_rev.assert_called_once_with({'id': 'rev'}, 10)
        mock_lookup_dir.assert_called_once_with('rev-id', 'some/path', True)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from collections import defaultdict
from datetime import datetime

from hypothesis import settings
from hypothesis.strategies import (
    just, sampled_from, lists, composite, datetimes
)
from string import ascii_letters, hexdigits

from swh.model.hashutil import hash_to_hex
from swh.storage.algos.revisions_walker import get_revisions_walker
from swh.storage.tests.algos.test_snapshot import origins
from swh.web.tests.data import get_tests_data

# Module dedicated to the generation of input data for tests through
# the use of hypothesis.
# Some of these data are sampled from a test archive created and populated
# in the swh.web.tests.data module.

# Set some hypothesis settings
settings.register_profile("swh-web", settings(deadline=None, max_examples=1))
settings.load_profile("swh-web")

# The following strategies exploit the hypothesis capabilities


def _known_swh_object(object_type):
    tests_data = get_tests_data()
    return sampled_from(tests_data[object_type])


def _unknown_swh_object(draw, object_type):
    tests_data = get_tests_data()
    while True:
        sha1_git = draw(sha1())
        if sha1_git not in tests_data[object_type]:
            return sha1_git


def sha1():
    """
    Hypothesis strategy returning a valid hexadecimal sha1 value.
    """
    sha1 = ''.join(random.choice(hexdigits) for x in range(40))
    return just(sha1.lower())


def invalid_sha1():
    """
    Hypothesis strategy returning an invalid sha1 representation.
    """
    invalid_sha1 = ''.join(random.choice(ascii_letters) for x in range(50))
    return just(invalid_sha1.lower())


def sha256():
    """
    Hypothesis strategy returning a valid hexadecimal sha256 value.
    """
    sha256 = ''.join(random.choice(hexdigits) for x in range(64))
    return just(sha256.lower())


def content():
    """
    Hypothesis strategy returning a random content ingested
    into the test archive.
    """
    return _known_swh_object('contents')


def contents():
    """
    Hypothesis strategy returning random contents ingested
    into the test archive.
    """
    return lists(content(), min_size=2, max_size=8)


@composite
def unknown_content(draw):
    """
    Hypothesis strategy returning a random content not ingested
    into the test archive.
    """
    tests_data = get_tests_data()
    while True:
        unknown_content = {
            'blake2s256': draw(sha256()),
            'sha1': draw(sha1()),
            'sha1_git': draw(sha1()),
            'sha256': draw(sha256())
        }
        if unknown_content not in tests_data['contents']:
            return unknown_content


def unknown_contents():
    """
    Hypothesis strategy returning random contents not ingested
    into the test archive.
    """
    return lists(unknown_content(), min_size=2, max_size=8)


def directory():
    """
    Hypothesis strategy returning a random directory ingested
    into the test archive.
    """
    return _known_swh_object('directories')


@composite
def unknown_directory(draw):
    """
    Hypothesis strategy returning a random directory not ingested
    into the test archive.
    """
    return _unknown_swh_object(draw, 'directories')


def origin():
    """
    Hypothesis strategy returning a random origin not ingested
    into the test archive.
    """
    return origins()


def visit_dates():
    """
    Hypothesis strategy returning a list of visit dates.
    """
    return lists(datetimes(min_value=datetime(2015, 1, 1, 0, 0),
                           max_value=datetime(2018, 12, 31, 0, 0)),
                 min_size=2, max_size=8, unique=True)


def release():
    """
    Hypothesis strategy returning a random release ingested
    into the test archive.
    """
    return _known_swh_object('releases')


@composite
def unknown_release(draw):
    """
    Hypothesis strategy returning a random revision not ingested
    into the test archive.
    """
    return _unknown_swh_object(draw, 'releases')


def revision():
    """
    Hypothesis strategy returning a random revision ingested
    into the test archive.
    """
    return _known_swh_object('revisions')


@composite
def unknown_revision(draw):
    """
    Hypothesis strategy returning a random revision not ingested
    into the test archive.
    """
    return _unknown_swh_object(draw, 'revisions')


def snapshot():
    """
    Hypothesis strategy returning a random snapshot ingested
    into the test archive.
    """
    return _known_swh_object('snapshots')


@composite
def unknown_snapshot(draw):
    """
    Hypothesis strategy returning a random revision not ingested
    into the test archive.
    """
    return _unknown_swh_object(draw, 'snapshots')


def _get_origin_dfs_revisions_walker():
    storage = get_tests_data()['storage']
    origin = random.choice(get_tests_data()['origins'])
    snapshot = storage.snapshot_get_latest(origin['id'])
    head = snapshot['branches'][b'HEAD']['target']
    return get_revisions_walker('dfs', storage, head)


def ancestor_revisions():
    """
    Hypothesis strategy returning a pair of revisions ingested into the
    test archive with an ancestor relation.
    """
    # get a dfs revisions walker for one of the origins
    # loaded into the test archive
    revisions_walker = _get_origin_dfs_revisions_walker()
    master_revisions = []
    children = defaultdict(list)
    init_rev_found = False
    # get revisions only authored in the master branch
    for rev in revisions_walker:
        for rev_p in rev['parents']:
            children[rev_p].append(rev['id'])
        if not init_rev_found:
            master_revisions.append(rev)
        if not rev['parents']:
            init_rev_found = True

    # head revision
    root_rev = master_revisions[0]
    # pick a random revision, different from head, only authored
    # in the master branch
    ancestor_rev_idx = random.choice(list(range(1, len(master_revisions)-1)))
    ancestor_rev = master_revisions[ancestor_rev_idx]
    ancestor_child_revs = children[ancestor_rev['id']]

    return just({
        'sha1_git_root': hash_to_hex(root_rev['id']),
        'sha1_git': hash_to_hex(ancestor_rev['id']),
        'children': [hash_to_hex(r) for r in ancestor_child_revs]
    })


def non_ancestor_revisions():
    """
    Hypothesis strategy returning a pair of revisions ingested into the
    test archive with no ancestor relation.
    """
    # get a dfs revisions walker for one of the origins
    # loaded into the test archive
    revisions_walker = _get_origin_dfs_revisions_walker()
    merge_revs = []
    children = defaultdict(list)
    # get all merge revisions
    for rev in revisions_walker:
        if len(rev['parents']) > 1:
            merge_revs.append(rev)
        for rev_p in rev['parents']:
            children[rev_p].append(rev['id'])
    # find a merge revisions whose parents have a unique child revision
    random.shuffle(merge_revs)
    selected_revs = None
    for merge_rev in merge_revs:
        if all(len(children[rev_p]) == 1
               for rev_p in merge_rev['parents']):
            selected_revs = merge_rev['parents']

    return just({
        'sha1_git_root': hash_to_hex(selected_revs[0]),
        'sha1_git': hash_to_hex(selected_revs[1])
    })

# The following strategies returns data specific to some tests
# that can not be generated and thus are hardcoded.


def contents_with_ctags():
    """
    Hypothesis strategy returning contents ingested into the test
    archive. Those contents are ctags compatible, that is running
    ctags on those lay results.
    """
    return just({
        'sha1s': ['0ab37c02043ebff946c1937523f60aadd0844351',
                  '15554cf7608dde6bfefac7e3d525596343a85b6f',
                  '2ce837f1489bdfb8faf3ebcc7e72421b5bea83bd',
                  '30acd0b47fc25e159e27a980102ddb1c4bea0b95',
                  '4f81f05aaea3efb981f9d90144f746d6b682285b',
                  '5153aa4b6e4455a62525bc4de38ed0ff6e7dd682',
                  '59d08bafa6a749110dfb65ba43a61963d5a5bf9f',
                  '7568285b2d7f31ae483ae71617bd3db873deaa2c',
                  '7ed3ee8e94ac52ba983dd7690bdc9ab7618247b4',
                  '8ed7ef2e7ff9ed845e10259d08e4145f1b3b5b03',
                  '9b3557f1ab4111c8607a4f2ea3c1e53c6992916c',
                  '9c20da07ed14dc4fcd3ca2b055af99b2598d8bdd',
                  'c20ceebd6ec6f7a19b5c3aebc512a12fbdc9234b',
                  'e89e55a12def4cd54d5bff58378a3b5119878eb7',
                  'e8c0654fe2d75ecd7e0b01bee8a8fc60a130097e',
                  'eb6595e559a1d34a2b41e8d4835e0e4f98a5d2b5'],
        'symbol_name': 'ABS'
    })



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.model.identifiers import REVISION

from swh.web.common.utils import reverse
from swh.web.common.exc import NotFoundExc
from swh.web.tests.testcase import WebTestCase


class SwhIdsApiTestCase(WebTestCase, APITestCase):

    @patch('swh.web.api.views.identifiers.service')
    def test_swh_id_resolve_success(self, mock_service):
        rev_id = '96db9023b881d7cd9f379b0c154650d6c108e9a3'
        origin = 'https://github.com/openssl/openssl'
        swh_id = 'swh:1:rev:%s;origin=%s' % (rev_id, origin)
        url = reverse('api-resolve-swh-pid', url_args={'swh_id': swh_id})

        mock_service.lookup_revision.return_value = {}

        resp = self.client.get(url)

        browse_rev_url = reverse('browse-revision',
                                 url_args={'sha1_git': rev_id},
                                 query_params={'origin': origin})

        expected_result = {
            'browse_url': browse_rev_url,
            'metadata': {'origin': origin},
            'namespace': 'swh',
            'object_id': rev_id,
            'object_type': REVISION,
            'scheme_version': 1
        }

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, expected_result)

    def test_swh_id_resolve_invalid(self):
        rev_id_invalid = '96db9023b8_foo_50d6c108e9a3'
        swh_id = 'swh:1:rev:%s' % rev_id_invalid
        url = reverse('api-resolve-swh-pid', url_args={'swh_id': swh_id})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 400)

    @patch('swh.web.api.views.identifiers.service')
    def test_swh_id_resolve_not_found(self, mock_service):
        rev_id_not_found = '56db90232881d7cd9e379b0c154650d6c108e9a1'

        swh_id = 'swh:1:rev:%s' % rev_id_not_found
        url = reverse('api-resolve-swh-pid', url_args={'swh_id': swh_id})

        mock_service.lookup_revision.side_effect = NotFoundExc('Revision not found !') # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 404)



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.test import APITestCase
from unittest.mock import patch

from swh.web.tests.testcase import WebTestCase


class PersonApiTestCase(WebTestCase, APITestCase):

    @patch('swh.web.api.views.person.service')
    def test_api_person(self, mock_service):
        # given
        stub_person = {
            'id': '198003',
            'name': 'Software Heritage',
            'email': 'robot@softwareheritage.org',
        }
        mock_service.lookup_person.return_value = stub_person

        # when
        rv = self.client.get('/api/1/person/198003/')

        # then
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, stub_person)

    @patch('swh.web.api.views.person.service')
    def test_api_person_not_found(self, mock_service):
        # given
        mock_service.lookup_person.return_value = None

        # when
        rv = self.client.get('/api/1/person/666/')

        # then
        self.assertEqual(rv.status_code, 404)
        self.assertEqual(rv['Content-Type'], 'application/json')
        self.assertEqual(rv.data, {
            'exception': 'NotFoundExc',
            'reason': 'Person with id 666 not found.'})



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64

from unittest.mock import patch

from django.utils.html import escape
from django.utils.encoding import DjangoUnicodeDecodeError

from swh.web.browse.utils import get_mimetype_and_encoding_for_content
from swh.web.common.exc import NotFoundExc
from swh.web.common.utils import reverse, get_swh_persistent_id
from swh.web.common.utils import gen_path_info
from swh.web.tests.testcase import WebTestCase

from .data.content_test_data import (
    stub_content_text_data,
    stub_content_text_path_with_root_dir,
    stub_content_bin_data,
    stub_content_bin_filename,
    stub_content_text_no_highlight_data,
    non_utf8_encoded_content_data,
    non_utf8_encoded_content,
    non_utf8_encoding,
    stub_content_too_large_data
)


class SwhBrowseContentTest(WebTestCase):

    @patch('swh.web.browse.views.content.request_content')
    def test_content_view_text(self, mock_request_content):
        mock_request_content.return_value = stub_content_text_data

        sha1_git = stub_content_text_data['checksums']['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}) # noqa

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': stub_content_text_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<code class="cpp">')
        self.assertContains(resp, escape(stub_content_text_data['raw_data']))
        self.assertContains(resp, url_raw)

        swh_cnt_id = get_swh_persistent_id('content', sha1_git)
        swh_cnt_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_cnt_id})
        self.assertContains(resp, swh_cnt_id)
        self.assertContains(resp, swh_cnt_id_url)

    @patch('swh.web.browse.views.content.request_content')
    def test_content_view_text_no_highlight(self, mock_request_content):
        mock_request_content.return_value = stub_content_text_no_highlight_data

        sha1_git = stub_content_text_no_highlight_data['checksums']['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_text_no_highlight_data['checksums']['sha1']}) # noqa

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': stub_content_text_no_highlight_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<code class="nohighlight">')
        self.assertContains(resp, escape(stub_content_text_no_highlight_data['raw_data'])) # noqa
        self.assertContains(resp, url_raw)

        swh_cnt_id = get_swh_persistent_id('content', sha1_git)
        swh_cnt_id_url = reverse('browse-swh-id',
                                 url_args={'swh_id': swh_cnt_id})

        self.assertContains(resp, swh_cnt_id)
        self.assertContains(resp, swh_cnt_id_url)

    @patch('swh.web.browse.utils.service')
    def test_content_view_no_utf8_text(self, mock_service):
        mock_service.lookup_content.return_value = \
            non_utf8_encoded_content_data

        mock_service.lookup_content_raw.return_value = \
            {'data': non_utf8_encoded_content}

        mock_service.lookup_content_filetype.return_value = None
        mock_service.lookup_content_language.return_value = None
        mock_service.lookup_content_license.return_value = None

        sha1_git = non_utf8_encoded_content_data['checksums']['sha1_git']

        url = reverse('browse-content',
                      url_args={'query_string': non_utf8_encoded_content_data['checksums']['sha1']}) # noqa

        try:
            resp = self.client.get(url)
            self.assertEqual(resp.status_code, 200)
            self.assertTemplateUsed('browse/content.html')
            swh_cnt_id = get_swh_persistent_id('content', sha1_git)
            swh_cnt_id_url = reverse('browse-swh-id',
                                     url_args={'swh_id': swh_cnt_id})
            self.assertContains(resp, swh_cnt_id_url)
            self.assertContains(resp, escape(non_utf8_encoded_content.decode(non_utf8_encoding).encode('utf-8'))) # noqa
        except DjangoUnicodeDecodeError:
            self.fail('Textual content is not encoded in utf-8')

    @patch('swh.web.browse.views.content.request_content')
    def test_content_view_image(self, mock_request_content):

        mime_type = 'image/png'
        mock_request_content.return_value = stub_content_bin_data

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_bin_data['checksums']['sha1']}) # noqa

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': stub_content_bin_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        png_encoded = base64.b64encode(stub_content_bin_data['raw_data']) \
                            .decode('utf-8')

        self.assertContains(resp, '<img src="data:%s;base64,%s"/>'
                                  % (mime_type, png_encoded))
        self.assertContains(resp, url_raw)

    @patch('swh.web.browse.views.content.request_content')
    def test_content_view_with_path(self, mock_request_content):
        mock_request_content.return_value = stub_content_text_data

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}, # noqa
                      query_params={'path': stub_content_text_path_with_root_dir}) # noqa

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, '<nav class="bread-crumbs')
        self.assertContains(resp, '<code class="cpp">')
        self.assertContains(resp, escape(stub_content_text_data['raw_data']))

        split_path = stub_content_text_path_with_root_dir.split('/')

        root_dir_sha1 = split_path[0]
        filename = split_path[-1]
        path = stub_content_text_path_with_root_dir \
            .replace(root_dir_sha1 + '/', '') \
            .replace(filename, '')

        path_info = gen_path_info(path)

        root_dir_url = reverse('browse-directory',
                               url_args={'sha1_git': root_dir_sha1})

        self.assertContains(resp, '<li class="swh-path">',
                            count=len(path_info)+1)

        self.assertContains(resp, '<a href="' + root_dir_url + '">' +
                            root_dir_sha1[:7] + '</a>')

        for p in path_info:
            dir_url = reverse('browse-directory',
                              url_args={'sha1_git': root_dir_sha1,
                                        'path': p['path']})
            self.assertContains(resp, '<a href="' + dir_url + '">' +
                                p['name'] + '</a>')

        self.assertContains(resp, '<li>' + filename + '</li>')

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': stub_content_text_data['checksums']['sha1']}, # noqa
                          query_params={'filename': filename})
        self.assertContains(resp, url_raw)

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}, # noqa
                      query_params={'path': filename})

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertNotContains(resp, '<nav class="bread-crumbs')

    @patch('swh.web.browse.views.content.request_content')
    def test_content_raw_text(self, mock_request_content):
        mock_request_content.return_value = stub_content_text_data

        url = reverse('browse-content-raw',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'text/plain')
        self.assertEqual(resp['Content-disposition'],
                         'filename=%s_%s' % ('sha1', stub_content_text_data['checksums']['sha1'])) # noqa
        self.assertEqual(resp.content, stub_content_text_data['raw_data'])

        filename = stub_content_text_path_with_root_dir.split('/')[-1]

        url = reverse('browse-content-raw',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}, # noqa
                      query_params={'filename': filename})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'text/plain')
        self.assertEqual(resp['Content-disposition'],
                         'filename=%s' % filename)
        self.assertEqual(resp.content, stub_content_text_data['raw_data'])

    @patch('swh.web.browse.utils.service')
    def test_content_raw_no_utf8_text(self, mock_service):
        mock_service.lookup_content.return_value = \
            non_utf8_encoded_content_data

        mock_service.lookup_content_raw.return_value = \
            {'data': non_utf8_encoded_content}

        mock_service.lookup_content_filetype.return_value = None
        mock_service.lookup_content_language.return_value = None
        mock_service.lookup_content_license.return_value = None

        url = reverse('browse-content-raw',
                      url_args={'query_string': non_utf8_encoded_content_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        _, encoding = get_mimetype_and_encoding_for_content(resp.content)
        self.assertEqual(encoding, non_utf8_encoding)

    @patch('swh.web.browse.views.content.request_content')
    def test_content_raw_bin(self, mock_request_content):
        mock_request_content.return_value = stub_content_bin_data

        url = reverse('browse-content-raw',
                      url_args={'query_string': stub_content_bin_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'application/octet-stream')
        self.assertEqual(resp['Content-disposition'],
                         'attachment; filename=%s_%s' %
                         ('sha1', stub_content_bin_data['checksums']['sha1']))
        self.assertEqual(resp.content, stub_content_bin_data['raw_data'])

        url = reverse('browse-content-raw',
                      url_args={'query_string': stub_content_bin_data['checksums']['sha1']}, # noqa
                      query_params={'filename': stub_content_bin_filename})

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp['Content-Type'], 'application/octet-stream')
        self.assertEqual(resp['Content-disposition'],
                         'attachment; filename=%s' % stub_content_bin_filename)
        self.assertEqual(resp.content, stub_content_bin_data['raw_data'])

    @patch('swh.web.browse.views.content.request_content')
    def test_content_request_errors(self, mock_request_content):

        url = reverse('browse-content', url_args={'query_string': '123456'})
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 400)
        self.assertTemplateUsed('error.html')

        mock_request_content.side_effect = NotFoundExc('content not found')

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_text_data['checksums']['sha1']}) # noqa
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 404)
        self.assertTemplateUsed('error.html')

    @patch('swh.web.browse.utils.service')
    def test_content_bytes_missing(self, mock_service):

        content_data = dict(stub_content_text_data)
        content_data['raw_data'] = None

        mock_service.lookup_content.return_value = content_data
        mock_service.lookup_content_raw.side_effect = NotFoundExc('Content bytes not available!') # noqa

        url = reverse('browse-content',
                      url_args={'query_string': content_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 404)
        self.assertTemplateUsed('browse/content.html')

    @patch('swh.web.browse.views.content.request_content')
    def test_content_too_large(self, mock_request_content):
        mock_request_content.return_value = stub_content_too_large_data

        url = reverse('browse-content',
                      url_args={'query_string': stub_content_too_large_data['checksums']['sha1']}) # noqa

        url_raw = reverse('browse-content-raw',
                          url_args={'query_string': stub_content_too_large_data['checksums']['sha1']}) # noqa

        resp = self.client.get(url)

        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed('browse/content.html')

        self.assertContains(resp, 'Content is too large to be displayed')
        self.assertContains(resp, url_raw)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.conf.urls import url


class UrlsIndex(object):
    """
    Simple helper class for centralizing url patterns of a Django
    web application.

    Derived classes should override the 'scope' class attribute otherwise
    all declared patterns will be grouped under the default one.
    """

    _urlpatterns = {}
    scope = 'default'

    @classmethod
    def add_url_pattern(cls, url_pattern, view, view_name):
        """
        Class method that adds an url pattern to the current scope.

        Args:
            url_pattern: regex describing a Django url
            view: function implementing the Django view
            view_name: name of the view used to reverse the url
        """
        if cls.scope not in cls._urlpatterns:
            cls._urlpatterns[cls.scope] = []
        if view_name:
            cls._urlpatterns[cls.scope].append(url(url_pattern, view,
                                                   name=view_name))
        else:
            cls._urlpatterns[cls.scope].append(url(url_pattern, view))

    @classmethod
    def get_url_patterns(cls):
        """
        Class method that returns the list of url pattern associated to
        the current scope.

        Returns:
            The list of url patterns associated to the current scope
        """
        return cls._urlpatterns[cls.scope]



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# flake8: noqa

import os

stub_content_root_dir = '08e8329257dad3a3ef7adea48aa6e576cd82de5b'

stub_content_text_file = \
"""
/* This file is part of the KDE project
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#ifndef KATE_SESSION_TEST_H
#define KATE_SESSION_TEST_H

#include <QObject>

class KateSessionTest : public QObject
{
    Q_OBJECT

private Q_SLOTS:
    void init();
    void cleanup();
    void initTestCase();
    void cleanupTestCase();

    void create();
    void createAnonymous();
    void createAnonymousFrom();
    void createFrom();
    void documents();
    void setFile();
    void setName();
    void timestamp();

private:
    class QTemporaryFile *m_tmpfile;
};

#endif
"""

stub_content_text_file_no_highlight = \
"""
                    GNU GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

                            Preamble

  The GNU General Public License is a free, copyleft license for
software and other kinds of works.

  The licenses for most software and other practical works are designed
to take away your freedom to share and change the works.  By contrast,
the GNU General Public License is intended to guarantee your freedom to
share and change all versions of a program--to make sure it remains free
software for all its users.  We, the Free Software Foundation, use the
GNU General Public License for most of our software; it applies also to
any other work released this way by its authors.  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
them if you wish), that you receive source code or can get it if you
want it, that you can change the software or use pieces of it in new
free programs, and that you know you can do these things.

  To protect your rights, we need to prevent others from denying you
these rights or asking you to surrender the rights.  Therefore, you have
certain responsibilities if you distribute copies of the software, or if
you modify it: responsibilities to respect the freedom of others.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must pass on to the recipients the same
freedoms that you received.  You must make sure that they, too, receive
or can get the source code.  And you must show them these terms so they
know their rights.

  Developers that use the GNU GPL protect your rights with two steps:
(1) assert copyright on the software, and (2) offer you this License
giving you legal permission to copy, distribute and/or modify it.

  For the developers' and authors' protection, the GPL clearly explains
that there is no warranty for this free software.  For both users' and
authors' sake, the GPL requires that modified versions be marked as
changed, so that their problems will not be attributed erroneously to
authors of previous versions.

  Some devices are designed to deny users access to install or run
modified versions of the software inside them, although the manufacturer
can do so.  This is fundamentally incompatible with the aim of
protecting users' freedom to change the software.  The systematic
pattern of such abuse occurs in the area of products for individuals to
use, which is precisely where it is most unacceptable.  Therefore, we
have designed this version of the GPL to prohibit the practice for those
products.  If such problems arise substantially in other domains, we
stand ready to extend this provision to those domains in future versions
of the GPL, as needed to protect the freedom of users.

  Finally, every program is threatened constantly by software patents.
States should not allow patents to restrict development and use of
software on general-purpose computers, but in those that do, we wish to
avoid the special danger that patents applied to a free program could
make it effectively proprietary.  To prevent this, the GPL assures that
patents cannot be used to render the program non-free.
"""

stub_content_text_data = {
    'checksums': {
        'sha1': '5ecd9f37b7a2d2e9980d201acd6286116f2ba1f1',
        'sha1_git': '537b47f68469c1c916c1bfbc072599133bfcbb21',
        'sha256': 'b3057544f04e5821ab0e2a007e2ceabd7de2dfb1d42a764f1de8d0d2eff80006',
        'blake2s256': '25117fa9f124d5b771a0a7dfca9c7a57247d81f8343334b4b41c782c7f7ed64d'
    },
    'length': 1317,
    'raw_data': str.encode(stub_content_text_file),
    'mimetype': 'text/x-c++',
    'encoding': 'us-ascii',
    'language': 'c++',
    'licenses': 'GPL',
    'error_code': 200,
    'error_message': '',
    'error_description': ''
}

stub_content_text_no_highlight_data = {
    'checksums': {
        'sha1': '8624bcdae55baeef00cd11d5dfcfa60f68710a02',
        'sha1_git': '94a9ed024d3859793618152ea559a168bbcbb5e2',
        'sha256': '8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b165a1dcd80c7c545eb65b903',
        'blake2s256': '38702b7168c7785bfe748b51b45d9856070ba90f9dc6d90f2ea75d4356411ffe'
    },
    'length': 35147,
    'raw_data': str.encode(stub_content_text_file_no_highlight),
    'mimetype': 'text/plain',
    'encoding': 'us-ascii',
    'language': 'not detected',
    'licenses': 'GPL',
    'error_code': 200,
    'error_message': '',
    'error_description': ''
}

stub_content_text_path = 'kate/autotests/session_test.h'

stub_content_text_path_with_root_dir = stub_content_root_dir + '/' + stub_content_text_path

stub_content_bin_filename = 'swh-logo.png'

png_file_path = os.path.dirname(__file__) + '/' + stub_content_bin_filename

with open(png_file_path, 'rb') as png_file:
    stub_content_bin_data = {
        'checksums': {
            'sha1': 'd0cec0fc2d795f0077c18d51578cdb228eaf6a99',
            'sha1_git': '02328b91cfad800e1d2808cfb379511b79679ebc',
            'sha256': 'e290592e2cfa9767497011bda4b7e273b4cf29e7695d72ecacbd723008a29144',
            'blake2s256': '7177cad95407952e362ee326a800a9d215ccd619fdbdb735bb51039be81ab9ce'
        },
        'length': 18063,
        'raw_data': png_file.read(),
        'mimetype': 'image/png',
        'encoding': 'binary',
        'language': 'not detected',
        'licenses': 'not detected',
        'error_code': 200,
        'error_message': '',
        'error_description': ''
    }

_non_utf8_encoding_file_path = os.path.dirname(__file__) + '/iso-8859-1_encoded_content'

non_utf8_encoded_content_data = {
    'checksums': {
        'sha1': '62cb71aa3534a03c12572157d20fa893753b03b6',
        'sha1_git': '2f7470d0b26108130e71087e42a53c032473499c',
        'sha256': 'aaf364ccd3acb546829ccc0e8e5e293e924c8a2e55a67cb739d249016e0034ed',
        'blake2s256': 'b7564932460a7c2697c53bd55bd855272490da511d64b20c5a04f636dc9ac467'
    },
    'length': 111000,
    'error_code': 200,
    'error_message': '',
    'error_description': ''
}

non_utf8_encoding = 'iso-8859-1'

with open(_non_utf8_encoding_file_path, 'rb') as iso88591_file:
    non_utf8_encoded_content = iso88591_file.read()

stub_content_too_large_data = {
    'checksums': {
        'sha1': '8624bcdae55baeef00cd11d5dfcfa60f68710a02',
        'sha1_git': '94a9ed024d3859793618152ea559a168bbcbb5e2',
        'sha256': '8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b165a1dcd80c7c545eb65b903',
        'blake2s256': '38702b7168c7785bfe748b51b45d9856070ba90f9dc6d90f2ea75d4356411ffe'
    },
    'length': 3000000,
    'raw_data': None,
    'mimetype': 'text/plain',
    'encoding': 'us-ascii',
    'language': 'not detected',
    'licenses': 'GPL',
    'error_code': 200,
    'error_message': '',
    'error_description': ''
}


# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.web.common import utils
from swh.web.common.exc import BadInputExc
from swh.web.tests.testcase import WebTestCase


class UtilsTestCase(WebTestCase):
    def test_shorten_path_noop(self):
        noops = [
            '/api/',
            '/browse/',
            '/content/symbol/foobar/'
        ]

        for noop in noops:
            self.assertEqual(
                utils.shorten_path(noop),
                noop
            )

    def test_shorten_path_sha1(self):
        sha1 = 'aafb16d69fd30ff58afdd69036a26047f3aebdc6'
        short_sha1 = sha1[:8] + '...'

        templates = [
            '/api/1/content/sha1:%s/',
            '/api/1/content/sha1_git:%s/',
            '/api/1/directory/%s/',
            '/api/1/content/sha1:%s/ctags/',
        ]

        for template in templates:
            self.assertEqual(
                utils.shorten_path(template % sha1),
                template % short_sha1
            )

    def test_shorten_path_sha256(self):
        sha256 = ('aafb16d69fd30ff58afdd69036a26047'
                  '213add102934013a014dfca031c41aef')
        short_sha256 = sha256[:8] + '...'

        templates = [
            '/api/1/content/sha256:%s/',
            '/api/1/directory/%s/',
            '/api/1/content/sha256:%s/filetype/',
        ]

        for template in templates:
            self.assertEqual(
                utils.shorten_path(template % sha256),
                template % short_sha256
            )

    def test_parse_timestamp(self):
        input_timestamps = [
            None,
            '2016-01-12',
            '2016-01-12T09:19:12+0100',
            'Today is January 1, 2047 at 8:21:00AM',
            '1452591542',
        ]

        output_dates = [
            None,
            datetime.datetime(2016, 1, 12, 0, 0),
            datetime.datetime(2016, 1, 12, 8, 19, 12,
                              tzinfo=datetime.timezone.utc),
            datetime.datetime(2047, 1, 1, 8, 21),
            datetime.datetime(2016, 1, 12, 9, 39, 2,
                              tzinfo=datetime.timezone.utc),
        ]

        for ts, exp_date in zip(input_timestamps, output_dates):
            self.assertEqual(utils.parse_timestamp(ts), exp_date)

    def test_format_utc_iso_date(self):
        self.assertEqual(utils.format_utc_iso_date('2017-05-04T13:27:13+02:00'), # noqa
                         '04 May 2017, 11:27 UTC')

    def test_gen_path_info(self):
        input_path = '/home/user/swh-environment/swh-web/'
        expected_result = [
            {'name': 'home', 'path': 'home'},
            {'name': 'user', 'path': 'home/user'},
            {'name': 'swh-environment', 'path': 'home/user/swh-environment'},
            {'name': 'swh-web', 'path': 'home/user/swh-environment/swh-web'}
        ]
        path_info = utils.gen_path_info(input_path)
        self.assertEqual(path_info, expected_result)

        input_path = 'home/user/swh-environment/swh-web'
        path_info = utils.gen_path_info(input_path)
        self.assertEqual(path_info, expected_result)

    def test_get_swh_persistent_id(self):
        swh_object_type = 'content'
        sha1_git = 'aafb16d69fd30ff58afdd69036a26047f3aebdc6'

        expected_swh_id = 'swh:1:cnt:' + sha1_git

        self.assertEqual(utils.get_swh_persistent_id(swh_object_type, sha1_git), # noqa
                         expected_swh_id)

        with self.assertRaises(BadInputExc) as cm:
            utils.get_swh_persistent_id('foo', sha1_git)
        self.assertIn('Invalid object', cm.exception.args[0])

        with self.assertRaises(BadInputExc) as cm:
            utils.get_swh_persistent_id(swh_object_type, 'not a valid id')
        self.assertIn('Invalid object', cm.exception.args[0])



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.web.common.utils import reverse
from swh.web.common.query import parse_hash


def filter_field_keys(data, field_keys):
    """Given an object instance (directory or list), and a csv field keys
    to filter on.

    Return the object instance with filtered keys.

    Note: Returns obj as is if it's an instance of types not in (dictionary,
    list)

    Args:
        - data: one object (dictionary, list...) to filter.
        - field_keys: csv or set of keys to filter the object on

    Returns:
        obj filtered on field_keys

    """
    if isinstance(data, map):
        return map(lambda x: filter_field_keys(x, field_keys), data)
    if isinstance(data, list):
        return [filter_field_keys(x, field_keys) for x in data]
    if isinstance(data, dict):
        return {k: v for (k, v) in data.items() if k in field_keys}
    return data


def person_to_string(person):
    """Map a person (person, committer, tagger, etc...) to a string.

    """
    return ''.join([person['name'], ' <', person['email'], '>'])


def enrich_object(object):
    """Enrich an object (revision, release) with link to the 'target' of
    type 'target_type'.

    Args:
        object: An object with target and target_type keys
        (e.g. release, revision)

    Returns:
        Object enriched with target_url pointing to the right
        swh.web.ui.api urls for the pointing object (revision,
        release, content, directory)

    """
    obj = object.copy()
    if 'target' in obj and 'target_type' in obj:
        if obj['target_type'] == 'revision':
            obj['target_url'] = reverse('api-revision',
                                        url_args={'sha1_git': obj['target']})
        elif obj['target_type'] == 'release':
            obj['target_url'] = reverse('api-release',
                                        url_args={'sha1_git': obj['target']})
        elif obj['target_type'] == 'content':
            obj['target_url'] = \
                reverse('api-content',
                        url_args={'q': 'sha1_git:' + obj['target']})

        elif obj['target_type'] == 'directory':
            obj['target_url'] = reverse('api-directory',
                                        url_args={'sha1_git': obj['target']})

    if 'author' in obj:
        author = obj['author']
        obj['author_url'] = reverse('api-person',
                                    url_args={'person_id': author['id']})

    return obj


enrich_release = enrich_object


def enrich_directory(directory, context_url=None):
    """Enrich directory with url to content or directory.

    """
    if 'type' in directory:
        target_type = directory['type']
        target = directory['target']
        if target_type == 'file':
            directory['target_url'] = \
                reverse('api-content', url_args={'q': 'sha1_git:%s' % target})
            if context_url:
                directory['file_url'] = context_url + directory['name'] + '/'
        elif target_type == 'dir':
            directory['target_url'] = reverse('api-directory',
                                              url_args={'sha1_git': target})
            if context_url:
                directory['dir_url'] = context_url + directory['name'] + '/'
        else:
            directory['target_url'] = reverse('api-revision',
                                              url_args={'sha1_git': target})
            if context_url:
                directory['rev_url'] = context_url + directory['name'] + '/'

    return directory


def enrich_metadata_endpoint(content):
    """Enrich metadata endpoint with link to the upper metadata endpoint.

    """
    c = content.copy()
    c['content_url'] = reverse('api-content',
                               url_args={'q': 'sha1:%s' % c['id']})
    return c


def enrich_content(content, top_url=False, query_string=None):
    """Enrich content with links to:
        - data_url: its raw data
        - filetype_url: its filetype information
        - language_url: its programming language information
        - license_url: its licensing information

    Args:
        content: dict of data associated to a swh content object
        top_url: whether or not to include the content url in
            the enriched data
        query_string: optional query string of type '<algo>:<hash>'
            used when requesting the content, it acts as a hint
            for picking the same hash method when computing
            the url listed above

    Returns:
        An enriched content dict filled with additional urls

    """
    checksums = content
    if 'checksums' in content:
        checksums = content['checksums']
    hash_algo = 'sha1'
    if query_string:
        hash_algo = parse_hash(query_string)[0]
    if hash_algo in checksums:
        q = '%s:%s' % (hash_algo, checksums[hash_algo])
        if top_url:
            content['content_url'] = reverse('api-content', url_args={'q': q})
        content['data_url'] = reverse('api-content-raw', url_args={'q': q})
        content['filetype_url'] = reverse('api-content-filetype',
                                          url_args={'q': q})
        content['language_url'] = reverse('api-content-language',
                                          url_args={'q': q})
        content['license_url'] = reverse('api-content-license',
                                         url_args={'q': q})

    return content


def _get_path_list(path_string):
    """Helper for enrich_revision: get a list of the sha1 id of the navigation
    breadcrumbs, ordered from the oldest to the most recent.

    Args:
        path_string: the path as a '/'-separated string

    Returns:
        The navigation context as a list of sha1 revision ids
    """
    return path_string.split('/')


def _get_revision_contexts(rev_id, context):
    """Helper for enrich_revision: retrieve for the revision id and potentially
    the navigation breadcrumbs the context to pass to parents and children of
    of the revision.

    Args:
        rev_id: the revision's sha1 id
        context: the current navigation context

    Returns:
        The context for parents, children and the url of the direct child as a
        tuple in that order.
    """
    context_for_parents = None
    context_for_children = None
    url_direct_child = None

    if not context:
        return (rev_id, None, None)

    path_list = _get_path_list(context)
    context_for_parents = '%s/%s' % (context, rev_id)
    prev_for_children = path_list[:-1]
    if len(prev_for_children) > 0:
        context_for_children = '/'.join(prev_for_children)
    child_id = path_list[-1]
    # This commit is not the first commit in the path
    if context_for_children:
        url_direct_child = reverse('api-revision-context',
                                   url_args={'sha1_git': child_id,
                                             'context': context_for_children})
    # This commit is the first commit in the path
    else:
        url_direct_child = reverse('api-revision',
                                   url_args={'sha1_git': child_id})

    return (context_for_parents, context_for_children, url_direct_child)


def _make_child_url(rev_children, context):
    """Helper for enrich_revision: retrieve the list of urls corresponding
    to the children of the current revision according to the navigation
    breadcrumbs.

    Args:
        rev_children: a list of revision id
        context: the '/'-separated navigation breadcrumbs

    Returns:
        the list of the children urls according to the context
    """
    children = []
    for child in rev_children:
        if context and child != _get_path_list(context)[-1]:
            children.append(reverse('api-revision',
                                    url_args={'sha1_git': child}))
        elif not context:
            children.append(reverse('api-revision',
                                    url_args={'sha1_git': child}))
    return children


def enrich_revision(revision, context=None):
    """Enrich revision with links where it makes sense (directory, parents).
    Keep track of the navigation breadcrumbs if they are specified.

    Args:
        revision: the revision as a dict
        context: the navigation breadcrumbs as a /-separated string of revision
        sha1_git
    """

    ctx_parents, ctx_children, url_direct_child = _get_revision_contexts(
        revision['id'], context)

    revision['url'] = reverse('api-revision',
                              url_args={'sha1_git': revision['id']})
    revision['history_url'] = reverse('api-revision-log',
                                      url_args={'sha1_git': revision['id']})
    if context:
        revision['history_context_url'] = reverse(
            'api-revision-log', url_args={'sha1_git': revision['id'],
                                          'prev_sha1s': context})

    if 'author' in revision:
        author = revision['author']
        revision['author_url'] = reverse('api-person',
                                         url_args={'person_id': author['id']})

    if 'committer' in revision:
        committer = revision['committer']
        revision['committer_url'] = \
            reverse('api-person', url_args={'person_id': committer['id']})

    if 'directory' in revision:
        revision['directory_url'] = \
            reverse('api-directory',
                    url_args={'sha1_git': revision['directory']})

    if 'parents' in revision:
        parents = []
        for parent in revision['parents']:
            parents.append({
                'id': parent,
                'url': reverse('api-revision', url_args={'sha1_git': parent})
            })

        revision['parents'] = parents

    if 'children' in revision:
        children = _make_child_url(revision['children'], context)
        if url_direct_child:
            children.append(url_direct_child)
        revision['children_urls'] = children
    else:
        if url_direct_child:
            revision['children_urls'] = [url_direct_child]

    if 'message_decoding_failed' in revision:
        revision['message_url'] = \
            reverse('api-revision-raw-message',
                    url_args={'sha1_git': revision['id']})

    return revision



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from bisect import bisect_right

from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

from swh.web import config
from swh.web.common import service
from swh.web.common.exc import BadInputExc, ForbiddenExc
from swh.web.common.models import (
    SaveUnauthorizedOrigin, SaveAuthorizedOrigin, SaveOriginRequest,
    SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_REJECTED, SAVE_REQUEST_PENDING,
    SAVE_TASK_NOT_YET_SCHEDULED, SAVE_TASK_SCHEDULED,
    SAVE_TASK_SUCCEED, SAVE_TASK_FAILED
)
from swh.web.common.origin_visits import get_origin_visits
from swh.web.common.utils import parse_timestamp

from swh.scheduler.utils import create_oneshot_task_dict

scheduler = config.scheduler()


def get_origin_save_authorized_urls():
    """
    Get the list of origin url prefixes authorized to be
    immediately loaded into the archive (whitelist).

    Returns:
        list: The list of authorized origin url prefix
    """
    return [origin.url
            for origin in SaveAuthorizedOrigin.objects.all()]


def get_origin_save_unauthorized_urls():
    """
    Get the list of origin url prefixes forbidden to be
    loaded into the archive (blacklist).

    Returns:
        list: the list of unauthorized origin url prefix
    """
    return [origin.url
            for origin in SaveUnauthorizedOrigin.objects.all()]


def can_save_origin(origin_url):
    """
    Check if a software origin can be saved into the archive.

    Based on the origin url, the save request will be either:

      * immediately accepted if the url is whitelisted
      * rejected if the url is blacklisted
      * put in pending state for manual review otherwise

    Args:
        origin_url (str): the software origin url to check

    Returns:
        str: the origin save request status, either **accepted**,
        **rejected** or **pending**
    """
    # origin url may be blacklisted
    for url_prefix in get_origin_save_unauthorized_urls():
        if origin_url.startswith(url_prefix):
            return SAVE_REQUEST_REJECTED

    # if the origin url is in the white list, it can be immediately saved
    for url_prefix in get_origin_save_authorized_urls():
        if origin_url.startswith(url_prefix):
            return SAVE_REQUEST_ACCEPTED

    # otherwise, the origin url needs to be manually verified
    return SAVE_REQUEST_PENDING


# map origin type to scheduler task
# TODO: do not hardcode the task name here
# TODO: unlock hg and svn loading once the scheduler
#       loading tasks are available in production
_origin_type_task = {
    'git': 'origin-update-git',
    # 'hg': 'origin-load-hg',
    # 'svn': 'origin-load-svn'
}


# map scheduler task status to origin save status
_save_task_status = {
    'next_run_not_scheduled': SAVE_TASK_NOT_YET_SCHEDULED,
    'next_run_scheduled': SAVE_TASK_SCHEDULED,
    'completed': SAVE_TASK_SUCCEED,
    'disabled': SAVE_TASK_FAILED
}


def get_savable_origin_types():
    return sorted(list(_origin_type_task.keys()))


def _check_origin_type_savable(origin_type):
    """
    Get the list of software origin types that can be loaded
    through a save request.

    Returns:
        list: the list of saveable origin types
    """
    allowed_origin_types = ', '.join(get_savable_origin_types())
    if origin_type not in _origin_type_task:
        raise BadInputExc('Origin of type %s can not be saved! '
                          'Allowed types are the following: %s' %
                          (origin_type, allowed_origin_types))


_validate_url = URLValidator(schemes=['http', 'https', 'svn', 'git'])


def _check_origin_url_valid(origin_url):
    try:
        _validate_url(origin_url)
    except ValidationError:
        raise BadInputExc('The provided origin url (%s) is not valid!' %
                          origin_url)


def _get_visit_date_for_save_request(save_request):
    visit_date = None
    try:
        origin = {'type': save_request.origin_type,
                  'url': save_request.origin_url}
        origin_info = service.lookup_origin(origin)
        origin_visits = get_origin_visits(origin_info)
        visit_dates = [parse_timestamp(v['date'])
                       for v in origin_visits]
        i = bisect_right(visit_dates, save_request.request_date)
        if i != len(visit_dates):
            visit_date = visit_dates[i]
    except Exception:
        pass
    return visit_date


def _save_request_dict(save_request, task=None):
    must_save = False
    visit_date = save_request.visit_date
    if task:
        save_task_status = _save_task_status[task['status']]
        if save_task_status in (SAVE_TASK_FAILED, SAVE_TASK_SUCCEED) \
                and not visit_date:
            visit_date = _get_visit_date_for_save_request(save_request)
            save_request.visit_date = visit_date
            must_save = True
        # Ensure last origin visit is available in database
        # before reporting the task execution as successful
        if save_task_status == SAVE_TASK_SUCCEED and not visit_date:
            save_task_status = SAVE_TASK_SCHEDULED
        if save_request.loading_task_status != save_task_status:
            save_request.loading_task_status = save_task_status
            must_save = True
        if must_save:
            save_request.save()
    else:
        save_task_status = save_request.loading_task_status

    return {'origin_type': save_request.origin_type,
            'origin_url': save_request.origin_url,
            'save_request_date': save_request.request_date.isoformat(),
            'save_request_status': save_request.status,
            'save_task_status': save_task_status,
            'visit_date': visit_date.isoformat() if visit_date else None}


def create_save_origin_request(origin_type, origin_url):
    """
    Create a loading task to save a software origin into the archive.

    This function aims to create a software origin loading task
    trough the use of the swh-scheduler component.

    First, some checks are performed to see if the origin type and
    url are valid but also if the the save request can be accepted.
    If those checks passed, the loading task is then created.
    Otherwise, the save request is put in pending or rejected state.

    All the submitted save requests are logged into the swh-web
    database to keep track of them.

    Args:
        origin_type (str): the type of origin to save (currently only
            ``git`` but ``svn`` and ``hg`` will soon be available)
        origin_url (str): the url of the origin to save

    Raises:
        BadInputExc: the origin type or url is invalid
        ForbiddenExc: the provided origin url is blacklisted

    Returns:
        dict: A dict describing the save request with the following keys:

            * **origin_type**: the type of the origin to save
            * **origin_url**: the url of the origin
            * **save_request_date**: the date the request was submitted
            * **save_request_status**: the request status, either **accepted**,
              **rejected** or **pending**
            * **save_task_status**: the origin loading task status, either
              **not created**, **not yet scheduled**, **scheduled**,
              **succeed** or **failed**


    """
    _check_origin_type_savable(origin_type)
    _check_origin_url_valid(origin_url)
    save_request_status = can_save_origin(origin_url)
    task = None

    # if the origin save request is accepted, create a scheduler
    # task to load it into the archive
    if save_request_status == SAVE_REQUEST_ACCEPTED:
        # create a task with high priority
        kwargs = {'priority': 'high'}
        # set task parameters according to the origin type
        if origin_type == 'git':
            kwargs['repo_url'] = origin_url
        elif origin_type == 'hg':
            kwargs['origin_url'] = origin_url
        elif origin_type == 'svn':
            kwargs['origin_url'] = origin_url
            kwargs['svn_url'] = origin_url

        sor = None
        # get list of previously sumitted save requests
        current_sors = \
            list(SaveOriginRequest.objects.filter(origin_type=origin_type,
                                                  origin_url=origin_url))

        can_create_task = False
        # if no save requests previously submitted, create the scheduler task
        if not current_sors:
            can_create_task = True
        else:
            # get the latest submitted save request
            sor = current_sors[0]
            # if it was in pending state, we need to create the scheduler task
            # and update the save request info in the database
            if sor.status == SAVE_REQUEST_PENDING:
                can_create_task = True
            # a task has already been created to load the origin
            elif sor.loading_task_id != -1:
                # get the scheduler task and its status
                tasks = scheduler.get_tasks([sor.loading_task_id])
                task = tasks[0] if tasks else None
                task_status = _save_request_dict(sor, task)['save_task_status']
                # create a new scheduler task only if the previous one has been
                # already executed
                if task_status == SAVE_TASK_FAILED or \
                   task_status == SAVE_TASK_SUCCEED:
                    can_create_task = True
                    sor = None
                else:
                    can_create_task = False

        if can_create_task:
            # effectively create the scheduler task
            task_dict = create_oneshot_task_dict(
                _origin_type_task[origin_type], **kwargs)
            task = scheduler.create_tasks([task_dict])[0]

            # pending save request has been accepted
            if sor:
                sor.status = SAVE_REQUEST_ACCEPTED
                sor.loading_task_id = task['id']
                sor.save()
            else:
                sor = SaveOriginRequest.objects.create(origin_type=origin_type,
                                                       origin_url=origin_url,
                                                       status=save_request_status, # noqa
                                                       loading_task_id=task['id']) # noqa
    # save request must be manually reviewed for acceptation
    elif save_request_status == SAVE_REQUEST_PENDING:
        # check if there is already such a save request already submitted,
        # no need to add it to the database in that case
        try:
            sor = SaveOriginRequest.objects.get(origin_type=origin_type,
                                                origin_url=origin_url,
                                                status=save_request_status)
        # if not add it to the database
        except ObjectDoesNotExist:
            sor = SaveOriginRequest.objects.create(origin_type=origin_type,
                                                   origin_url=origin_url,
                                                   status=save_request_status)
    # origin can not be saved as its url is blacklisted,
    # log the request to the database anyway
    else:
        sor = SaveOriginRequest.objects.create(origin_type=origin_type,
                                               origin_url=origin_url,
                                               status=save_request_status)

    if save_request_status == SAVE_REQUEST_REJECTED:
        raise ForbiddenExc('The origin url is blacklisted and will not be '
                           'loaded into the archive.')

    return _save_request_dict(sor, task)


def get_save_origin_requests_from_queryset(requests_queryset):
    """
    Get all save requests from a SaveOriginRequest queryset.

    Args:
        requests_queryset (django.db.models.QuerySet): input
            SaveOriginRequest queryset

    Returns:
        list: A list of save origin requests dict as described in
        :func:`swh.web.common.origin_save.create_save_origin_request`
    """
    task_ids = []
    for sor in requests_queryset:
        task_ids.append(sor.loading_task_id)
    tasks = scheduler.get_tasks(task_ids)
    tasks = {task['id']: task for task in tasks}
    requests = []
    for sor in requests_queryset:
        sr_dict = _save_request_dict(sor, tasks.get(sor.loading_task_id, None))
        requests.append(sr_dict)
    return requests


def get_save_origin_requests(origin_type, origin_url):
    """
    Get all save requests for a given software origin.

    Args:
        origin_type (str): the type of the origin
        origin_url (str): the url of the origin

    Raises:
        BadInputExc: the origin type or url is invalid

    Returns:
        list: A list of save origin requests dict as described in
        :func:`swh.web.common.origin_save.create_save_origin_request`
    """
    _check_origin_type_savable(origin_type)
    _check_origin_url_valid(origin_url)
    sors = SaveOriginRequest.objects.filter(origin_type=origin_type,
                                            origin_url=origin_url)
    return get_save_origin_requests_from_queryset(sors)
